'/* BROWSE.BAS  Multi-file Btrieve/ISAM/etc. Browse/select module */
'/*             By: Dale Thorn                                    */
'/*             Rev. 21.11.2002                                   */

'  IN:  structure bc..  bc.auxkey - 0 = display & exit (position may exist)
'                                   1 = reentry w/o display (existing position)
'                                 > 1 = normal entry (position may exist)
'
'  Note:  If ckey(0) = NULLS, do HOME call, else do GFPOS.
'         If cpos(0) = NULLS, get Btrieve position, else use cpos(0) position.
'
'                       bc.clrbar - scroll-bar color
'                       bc.clrdat - data area color
'                       bc.clredt - edit-cell color
'                       bc.clrerr - error message color
'                       bc.clrftr - footer message(s) color
'                       bc.clrhdr - main header title color
'                       bc.clrmsg - standard message color
'                       bc.clrttl - column titles color
'                       bc.clrwin - window border color
'                       bc.colmin - window start column (-1 = autoconfig)
'                       bc.declen - no. decimal places for float-type fields
'                       bc.edtmax - no. of search fields - 1
'                       bc.filmax - no. of files accessed by this module - 1
'                       bc.fldmax - no. of fields on window - 1
'                       bc.fldmin - begin field no. (poss. neg. & not displayed)
'                       bc.holdln - TRUE if OK to hold one line on paging
'                       bc.keyout - buffered-key timeout (0 = no timeout)
'                       bc.linext - window-border line width (0, 1 or 2)
'                       bc.linint - internal-line width (0 or 1)
'                       bc.padext - TRUE if OK to pad l & r of screen only
'                       bc.padint - TRUE if OK to pad all data fields
'                       bc.rowmin - window start row    (-1 = autoconfig)
'                       bc.rowtot - window size in rows (-1 = autoconfig)
'                       bc.xkey01 to 03:  Exit-Browse key values
'         cdat(0,n) - (0) the screen header title
'                     (1) subheader if possible to select NEW record
'                     (2) optional message:  <records/data/info> not found etc.
'                     (3) first  footer message
'                     (4) second footer message
'                     (5) third  footer message
'         cdat(1,n) - (0-22) field formatting buffers (set internally)
'         cdat(2,n) - (0-22) field (column) header titles (1st row)
'         cdat(3,n) - (0-22) field (column) header titles (2nd row)
'         cdat(4,n) - (0-22) field (column) types:
'                       A1 = ASCII byte (0 to 255)
'                       C1 = character (ASCII 32 to 127)
'                       C2 = character (ASCII 32 to 127, variable length)
'                       D1 = date in mmddyy or mmddyyyy format
'                       D2 = date in yymmdd or yyyymmdd format
'                       D5 = date in dmyy format (Btrieve)
'                       E1 = expandable character field (phone, zip, ...)
'                       H6 = extendable right-just.float [**] (Double IEEE)
'                       H7 = extendable right-just.float [*/] (Double IEEE)
'                       H8 = extendable right-just.float [/*] (Double IEEE)
'                       H9 = extendable right-just.float [//] (Double IEEE)
'                       I5 = right-justified float numeric (Single IEEE)
'                       I6 = same as type I5 except displayed zero-filled
'                       I7 = right-justified float numeric (Double IEEE)
'                       I8 = same as type I7 except displayed zero-filled
'                       M6 = extendable right-just.float [**] (Double MBF)
'                       M7 = extendable right-just.float [*/] (Double MBF)
'                       M8 = extendable right-just.float [/*] (Double MBF)
'                       M9 = extendable right-just.float [//] (Double MBF)
'                       N0 = right-justified dollar$ numeric (Double MBF)
'                       N1 = right-justified integer numeric (Short integer)
'                       N2 = same as type N1 except displayed zero-filled
'                       N3 = right-justified integer numeric (Long integer)
'                       N4 = same as type N3 except displayed zero-filled
'                       N5 = right-justified float numeric (Single MBF)
'                       N6 = same as type N5 except displayed zero-filled
'                       N7 = right-justified float numeric (Double MBF)
'                       N8 = same as type N7 except displayed zero-filled
'                       N9 = right-justified float numeric (Character)
'                       S0 = replaceable field type (source index base = 0)
'                       S1 = replaceable field type (source index base = 1)
'                       S2 = updateable field type  (source index base = 0)
'                       S3 = updateable field type  (source index base = 1)
'                       S4 = updateable field type  (source lookup)
'                       S5 = replaceable field type (source lookup)
'                       S6 = replaceable field type (constant substitution)
'                       T1 = time in hhmm format (hhmm or hhmmss)
'                       T5 = time in ssmh format (Btrieve)
'         cdat(5,n) - ( 0-22)
'                     field-segment separator chars. ("E1" type fields only)
'                     extension-field number ("M6" thru "M9" type fields only)
'                     field source (lookup) strings ("S4"&"S5" type fields only)
'         cdat(6,n) - ( 0-22)
'                     ASCII field-segment lengths ("E1" type fields only)
'                     extension-field constant ("M6" thru "M9" type fields only)
'                     field replacemnt.strings ("S0" thru "S5" type fields only)
'         cdat(7,n) - ( 0- 9) search field buffers (set internally)
'                     (10-19) left-pad default data for Btrieve key buffers
'         ckey()    - ( 0- 9) file access key buffers
'                     (10-19) masking for ckey(0-9)  [CCCXXXII...]
'               ***** NOTE: masking characters must be uppercased *****
'
'                     (20) minimum actual key value possible for this file
'                     (21) maximum actual key value possible for this file
'                     (22-28) key-comparison buffers (set internally)
'         ckvl()    - ( 0-22) key-value data-row buffers (set internally)
'         cpos()    - ( 0-27) Btrv. data-row position buffers (set internally)
'         crcd(0,n) - ( 0- 9) data-record buffers
'         crcd(1,n) - ( 0- 9) file-control-block buffers
'         cvdb()    - ( 0-27) video data-row buffers (set internally)
'         iptr(0,n) - ( 0-22) beginning byte position of data fields in crcd()
'         iptr(1,n) - ( 0-22) internal length of data fields in crcd()
'         iptr(2,n) - ( 0-22) screen width of each data field (-1 = autoconfig)
'         iptr(3,n) - ( 0-22) file# pointers for each field (file #s 0-9)
'         iptr(4,n) - ( 0- 9) Btrieve file record lengths
'                     (10-19) unused
'         iptr(5,n) - ( 0- 9) search-field field no. (-1 = not appl.here&below)
'                     (10-19) search field start column relative to window
'         iptr(6,n) - ( 0- 9) search field length       (-1 = not appl.)
'                     (10-19) search field end column +1 (set internally)
'         iptr(7,n) - ( 0- 9) Btrieve file access key numbers
'                     (10) no. of data segments (-1) to load from file #0
'                          key** buffer for use in search code [case 32-127]
'                     (11-19) no. of data segments (-1) to load from file #0
'                             data** buffer into key buffers #1 and up
'         iptr(n,n) - begin and length pointers for data segments in iptr(7, n)
'
'                  ** pointers for file #0 are from key (not record) buffer **
'                  ** and are used only in the search code [case 32 to 127] **
'
'
' OUT:    ckey(0)   - NEW selected or no records = NULL, else current key value
'       bc.auxkey  (returned in first 2 bytes of bc structure)
'                     -1 = zero records, else "quit" key from calling program
'==============================================================================

'$include: 'basdef.h'
'$include: 'basio.h'
'$include: 'btrieve.h'
'$include: 'browse.h'
'$include: 'filekill.h'
'$include: 'fileopen.h'
'$include: 'longname.h'
'$include: 'messages.h'
'$include: 'midchar.h'
'$include: 'parmstr1.h'
'$include: 'string.h'

function browse(bc as bcnf, cdat(), ckey(), ckvl(), cpos(), crcd(), cvdb(), iptr())
   bc.chradj = bc.linext - abs(bc.linext > 0)'used to set line-width characters
   bc.padadj = abs(bc.padint) * 2                      'internal pad amount * 2
   if bc.padint then                         'OK to pad internal display fields
      bc.padext = 0                    'no external pad when internal pad is on
   end if
   icoltot = 1 + abs(bc.padext) * 2       'initialize total Browse window width
   itagfld = -1                          'initialize the replaceable field type
   bc.titles = 0                      'initialize the field (column) title rows
   for ifld = bc.fldmin to bc.fldmax       'loop thru the Browse display fields
      if iptr(2, ifld) = -1 then                  'if width = -1, do autoconfig
         select case cdat(4, ifld)                   'select display field type
            case "A1"                                    'ASCII byte (0 to 255)
               iptr(2, ifld) = 3                   'set field display width = 3
            case "C1", _                                        'character type
                 "C2", _                      'character type (variable length)
                 "N9", _                      'numeric type stored as character
                 "S0", _                    'replaceable type (begin-index = 0)
                 "S1", _                    'replaceable type (begin-index = 1)
                 "S2", _                     'updateable type (begin-index = 0)
                 "S3", _                     'updateable type (begin-index = 1)
                 "S4", _                       'updateable type (source lookup)
                 "S5", _                      'replaceable type (source lookup)
                 "S6"                         'replaceable type (const. subst.)
               iptr(2, ifld) = iptr(1, ifld) 'set to internal record field size
            case "D1", _                          'date type mmddyy or mmddyyyy
                 "D2", _                          'date type yymmdd or yyyymmdd
                 "D5", _                              'date type dmyy (Btrieve)
                 "I5", _                        'Single IEEE float numeric type
                 "I6", _                           'same as I5 plus zero-filled
                 "N3", _                             'Long integer numeric type
                 "N4", _                           'same as N3 plus zero-filled
                 "N5", _                         'Single MBF float numeric type
                 "N6"                              'same as N5 plus zero-filled
               iptr(2, ifld) = 10                 'set field display width = 10
            case "E1"                         'expandable char.(phone, zip,...)
               iptr(2, ifld) = iptr(1, ifld) + len(cdat(5, ifld))'fldsz+no.seps
            case "H6", _                      'Double IEEE extendable [**] type
                 "H7", _                      'Double IEEE extendable [*/] type
                 "H8", _                      'Double IEEE extendable [/*] type
                 "H9", _                      'Double IEEE extendable [//] type
                 "I7", _                        'Double IEEE float numeric type
                 "I8", _                           'same as I7 plus zero-filled
                 "M6", _                       'Double MBF extendable [**] type
                 "M7", _                       'Double MBF extendable [*/] type
                 "M8", _                       'Double MBF extendable [/*] type
                 "M9", _                       'Double MBF extendable [//] type
                 "N0", _                        'Double MBF dollar & cents type
                 "N7", _                         'Double MBF float numeric type
                 "N8"                              'same as N7 plus zero-filled
               iptr(2, ifld) = 17                 'set field display width = 17
            case "N1", _                            'Short integer numeric type
                 "N2"                              'same as N1 plus zero-filled
               iptr(2, ifld) = 5                   'set field display width = 5
            case "T1", _                              'time type hhmm or hhmmss
                 "T5"                                 'time type ssmh (Btrieve)
               iptr(2, ifld) = 8                   'set field display width = 8
            case else
         end select
      end if
      iextlen = iptr(2, ifld) + bc.padadj           'padded display-field width
      if icoltot + iextlen < 80 then        'total width OK; process this field
         if asc(cdat(4, ifld)) = 83 then             '"S"-type selection fields
            iseltyp = midchar(cdat(4, ifld), 2) - 48  'select-field type number
            if iseltyp <= 3 then        'source selection-string not applicable
               iselcnt = len(cdat(6, ifld)) \ iptr(2, ifld)  'no. of selections
               cdat(5, ifld) = space$(iselcnt * iptr(1, ifld))
            end if              'set up a dummy source selection-string [above]
            if iseltyp >= 2 and iseltyp <= 4 and ifld >= 0 then  'toggle fields
               itagcol = icoltot - abs(bc.padext) + abs(bc.padint)'begin column
               itagfld = ifld         'identifies single allowable toggle field
            end if
         end if
         for irow = 2 to 3                 'loop thru column (field) title rows
            if len(cdat(irow, ifld)) then      'title(s) were supplied this row
               if len(cdat(irow, ifld)) > iextlen then  'title wider than field
                  cdat(irow, ifld) = left$(cdat(irow, ifld), iextlen)'trim data
               end if
               if bc.titles < irow then         'first title found for this row
                  bc.titles = irow                  'increment total title rows
               end if
            end if
         next
         if ifld >= 0 then                                   'displayable field
            icoltot = icoltot + iextlen + 1     'accumulated total window width
         end if
         iptr(2, ifld) = iextlen         'put padded field width in width array
      else                              'max. width exceeded; terminate process
         bc.fldmax = ifld - 1               'decrement #fields to current total
         exit for                          'terminate Browse display field loop
      end if
   next
   if bc.rowtot = -1 then                       'total rows = -1; do autoconfig
      bc.rowtot = 25                          'set row total = 25 (full screen)
   end if
   if bc.colmin = -1 then                  'start column = -1; do autocentering
      bc.colmin = (80 - icoltot) \ 2        'set begin column for autocentering
   end if
   if bc.rowmin = -1 then                     'start row = -1; do autocentering
      bc.rowmin = (25 - bc.rowtot) \ 2         'set begin row for autocentering
   end if
   bc.colmax = bc.colmin + icoltot - 1                'Browse window end column
   bc.colcnt = icoltot - 2                      'total length of displayed data
   for indx = 0 to bc.edtmax                'loop thru the Browse search fields
      iptr(5, indx + 10) = iptr(5, indx + 10) + bc.colmin  'adjust begin column
      iptr(6, indx + 10) = iptr(5, indx + 10) + iptr(6, indx)    'end column +1
   next
   bc.ftrcnt = 0                               'initialize footer-message count
   iftrlen = 0                           'initialize max. footer-message length
   for indx = 3 to 5                        'loop thru the footer-message array
      if len(cdat(0, indx)) then                'a footer was supplied this row
         bc.ftrcnt = bc.ftrcnt + 1              'increment footer-message count
      end if
      if iftrlen < len(cdat(0, indx)) then       'current footer is the longest
         iftrlen = len(cdat(0, indx))         'reset max. footer-message length
      end if
   next
   bc.ftrcol = bc.colmin + (icoltot - iftrlen) \ 2     'set footer start column
   bc.hdrcol = bc.colmin + (icoltot - len(cdat(0, 0))) \ 2 'header start column
   bc.hdrpos = (82 - len(cdat(0, 0))) \ 2      'header start column for reports
   bc.hdrrow = (left$(cdat(0, 0), 1) > " ")  'TRUE if 1st char. of header > ' '
   inewrcd = (len(cdat(0, 1)) > 0)       'TRUE if possible to select NEW record
   imaxlen = 0              'initialize max. key length for related-record gets
   for indx = 0 to bc.filmax               'loop thru #files accessed by Browse
      if imaxlen < len(ckey(indx)) then          'current key length is longest
         imaxlen = len(ckey(indx))                    'reset maximum key length
      end if
   next
   bc.rowmax = bc.rowmin + bc.rowtot - 1                 'Browse window end row
   bc.rowbeg = bc.rowmin + abs(bc.hdrrow) * 2 + bc.titles + 1   'first data row
   bc.rowend = bc.rowmax - abs(bc.edtmax >= 0) _                 'last data row
   * 2 - abs(bc.ftrcnt > 0) - bc.ftrcnt - 1
   bc.rowcnt = bc.rowend - bc.rowbeg + 1         'total Browse window data rows
   iedtrow = bc.rowend + 2                      'Browse window search field row
   ikeylen = len(ckey(0))                       'length of main file-access key
   if bc.auxkey <> 1 then                  'OK to display initial Browse window
      bc.dsprow = -1                          'initialize current (display) row
      bc.cdtl = char(45)                   'bottom-left connector in header box
      bc.cdtr = char(45)                  'bottom-right connector in header box
      bc.cllt = char(92)                   'bottom-left corner of Browse window
      bc.clrt = char(47)                  'bottom-right corner of Browse window
      bc.cstd = char(45)                   'down-side connector for int'l lines
      bc.cstl = char(124)                  'left-side connector for int'l lines
      bc.cstr = char(124)                 'right-side connector for int'l lines
      bc.cstu = char(45)                     'up-side connector for int'l lines
      for indx = 22 to 27           'loop thru key buffers used for comparisons
         ckey(indx) = string$(ikeylen, 0)   'initialize the key buffer to NULLs
      next
      ckey(28) = string$(imaxlen, 0)        'key-buffer for related-record gets
      for ifld = bc.fldmin to bc.fldmax    'loop thru the Browse display fields
         cdat(1, ifld) = space$(iptr(2, ifld) - bc.padadj)'field format buffers
      next
      for indx = 0 to bc.edtmax             'loop thru the Browse search fields
         cdat(7, indx) = space$(iptr(6, indx))   'allocate search-field buffers
      next
      ckvl(0) = string$(ikeylen, 255)     'initial value of 1st-row data buffer
      for irow = 1 to bc.rowcnt - 1      'loop thru the Browse window data rows
         ckvl(irow) = string$(ikeylen, 0)  'initialize the data row key buffers
      next
      for irow = 1 to 27                   'loop thru the position-buffer array
         cpos(irow) = string$(4, 0)    'initialize the Btrieve position buffers
      next
      for irow = 0 to 24                     'loop thru the video display array
         cvdb(irow) = space$(bc.colcnt)       'allocate the video display array
      next
      cvdb(25) = space$(768)                                 'print-line buffer
      cvdb(26) = string$(bc.colcnt, 45)                   'Browse window border
      cvdb(27) = space$(25)        'buffer used in formatting float-type fields
      i = ibr.scrn(bc, cdat(), cvdb(), iptr())       'display the Browse window
   end if
   if ckey(0) = ckey(26) then              'incoming key = NULLs (do HOME call)
      ikey = KEY.HOME                          '"user" key for HOME positioning
      cpos(0) = space$(4)                  'make position-buffer #0 <> cpos(23)
   else                                  'incoming key <> NULLs (do GFPOS call)
      ikey = KEY.GFPOS                        '"user" key for Get-From-POSition
      if len(cpos(0)) then        'position-buffer #0 contains a valid position
         lset cpos(25) = cpos(0)     'set comparison pos.-buffer for GFPOS call
      else                              'position-buffer #0 NOT yet initialized
         cpos(0) = cpos(26)             'initialize position-buffer #0 to NULLS
      end if
   end if
   bc.dspnew = 0                            'regular or NEW record display flag
   bc.prtque = 0                                       'printing is active flag
   iprtset = 0                                           'print mode is ON flag
   ibufkey = 0                           'initialize keyboard buffered-key flag
   ibypass = (bc.auxkey <> 1)   'set keyboard-bypass ON for initial positioning
   iedtcol = iptr(5, 10)                'current position of search-cell cursor
   iedtfld = 0                                       'current edit-field number
   do                                                   'begin main Browse loop
      if ibypass then                    'automatic (non-user) key-select is ON
         ibypass = 0                     'reset automatic (non-user) key-select
      else                              'automatic (non-user) key-select is OFF
         if bc.edtmax >= 0 then                  'search field(s) are available
            i = io.vcsr(iedtrow, iedtcol, ICSR.COVR)'turn cursor ON for key-get
         else                                'search field(s) are NOT available
            i = io.vcsr(bc.rowmax, bc.colmax, 0)               'turn cursor OFF
         end if
         if ibufkey then                 'buffered-key state for searches is ON
            ikey = ibr.kbuf(bc)       'get the next buffered key (wait if none)
         else                           'buffered-key state for searches is OFF
            ikey = io.kget(0)       'get the next unbuffered key (wait for key)
         end if
         if bc.edtmax >= 0 then                  'search field(s) are available
            i = io.vcsr(bc.rowmax, bc.colmax, 0) 'turn cursor OFF after key-get
         end if
         if bc.edtmax > 0 and ikey = KEY.TABRT then    'Tab key in search field
            ikey = KEY.CSRRT                    'change Tab key to right-cursor
         end if
         if (ikey < 32 and ikey <> KEY.TABRT or ikey > 127) _
         and ikey <> KEY.CSRRT and ibufkey then      'OK to clear search fields
            ibufkey = 0                             'turn buffered-key flag OFF
            iedtcol = iptr(5, 10)       'set current search column to beginning
            iedtfld = 0                  'set current search field to beginning
            for indx = 0 to bc.edtmax              'loop thru the search fields
               lset cdat(7, indx) = ""          'clear the search field buffers
               i = io.vdsp(cdat(7, indx), iedtrow, _
               iptr(5, indx + 10), bc.clrwin)    'clear search fields on window
            next
         end if
      end if
      select case ikey                    'select user-hit or default key value
         case KEY.CTEND, _               'Ctrl-End (display last 2 fields only)
              KEY.CTHOM, _              'Ctrl-Home (begin display at 1st field)
              KEY.CTLLT, _            'Ctrl-Left (move display 1 field to left)
              KEY.CTLRT             'Ctrl-Right (move display 1 field to right)
            exit do                                           'exit-from-Browse
         case KEY.CSRDN                                        'Down-cursor key
            i = ibr.hkey(ckey(), ckvl(), cpos(), crcd(), iptr())  'get high key
            if cpos(0) <> cpos(24) and cpos(24) <> cpos(26) or bc.dsprow > 0 then
               bc.scroll = -1                     'set scroll flag to scroll up
               istat = ibr.btrv(BTRV.GETDIRECT, crcd(1, 0), crcd(0, 0), _
               iptr(4, 0), ckey(0), iptr(7, 0), cpos(bc.dsprow))'get f/saved pos
               istat = ibr.btrv(BTRV.GETNEXT, crcd(1, 0), crcd(0, 0), _
               iptr(4, 0), ckey(0), iptr(7, 0), cpos(25))  'get the next record
               i = ibr.mask(ckey(), ckey(0), ckey(21))       'mask the file key
               if istat = 0 and ckey(24) <= ckey(25) then   'valid record found
                  i = ibr.scrl(bc, ckvl(), cpos(), cvdb())'scroll data up 1 row
                  i = ibr.disp(bc, cdat(), ckey(), crcd(), cvdb(), iptr())
                  lset ckvl(bc.dsprow) = ckey(0)            'save new key value
                  lset cpos(bc.dsprow) = cpos(25)     'save new position buffer
               else                                     'valid record NOT found
                  for indx = bc.dsprow to 0 step -1    'loop thru the data rows
                     lset cpos(25) = cpos(indx)    'see if record still in file
                     istat = ibr.btrv(BTRV.GETDIRECT, crcd(1, 0), crcd(0, 0), _
                     iptr(4, 0), ckey(0), iptr(7, 0), cpos(25))   'get key data
                     if istat = 0 then                 'record IS still in file
                        exit for                    'stop looking at this point
                     end if
                  next
                  if indx >= 0 then             'next key not found; clear data
                     if indx > 0 then           'record found is NOT 1st record
                        i = ibr.scrl(bc, ckvl(), cpos(), cvdb())'scroll rows up
                     end if
                     if indx = bc.dsprow then   'valid record found at last row
                        bc.dsprow = bc.dsprow - 1           'decrement last row
                     else                 'valid record found prior to last row
                        bc.dsprow = indx - abs(indx > 0)  'reset last disp. row
                     end if    'clear all data following last valid row [below]
                     i = ibr.cdat(bc, ckey(), ckvl(), cpos(), cvdb())
                  else                    'no valid data on screen; do END call
                     ikey = KEY.END                      'set "user" key to END
                     ibypass = not 0                 'set keyboard bypass to ON
                  end if
               end if
            else                          'no records available for down-cursor
               if cpos(24) = cpos(26) then  'no records available in data range
                  ikey = KEY.HOME                       'set "user" key to HOME
                  ibypass = not 0                    'set keyboard bypass to ON
               end if
            end if
         case KEY.CSRRT                                       'Right-cursor key
            if iedtfld < bc.edtmax and cpos(0) <> cpos(26) then'OK to jump-edit
               ibufkey = not 0                    'set buffered key-get flag ON
               lset ckey(0) = ckvl(0)         'get data record with current key
               istat = ibr.btrv(BTRV.GETEQ, crcd(1, 0), crcd(0, 0), _
               iptr(4, 0), ckey(0), iptr(7, 0), cpos(25))       'get the record
               if istat then                                  'record NOT found
                  lset cdat(7, iedtfld) = ""            'clear the search field
               else                                       'the record was found
                  ifld = iptr(5, iedtfld)    'temp. copy of search-field number
                  if cdat(4, ifld) = "E1" then       'search field type is "E1"
                     csep = cdat(5, ifld) + cdat(6, ifld) 'sep.+segment lengths
                  else                  'search-field types are other than "E1"
                     csep = bc.csep                   'separator character only
                  end if                         'format the edit-field [below]
                  i = ibr.fmtx(mid$(crcd(0, 0), iptr(0, ifld), iptr(1, ifld)), _
                  cdat(7, iedtfld), cdat(4, ifld), csep, iptr(6, iedtfld))
               end if
               i = io.vdsp(cdat(7, iedtfld), iedtrow, _
               iptr(5, iedtfld + 10), bc.clredt)        'display the edit-field
               iedtfld = iedtfld + 1                  'increment the edit-field
               lset cdat(7, iedtfld) = ""            'clear the next edit-field
               iedtcol = iptr(5, iedtfld + 10)   'set edit-column to next field
               i = io.vdsp(cdat(7, iedtfld), iedtrow, iedtcol, bc.clredt)
            end if
         case KEY.CSRUP                                          'Up-cursor key
            i = ibr.lkey(ckey(), ckvl(), cpos(), crcd(), iptr())   'get low key
            if cpos(0) <> cpos(26) and (cpos(0) <> cpos(23) or inewrcd) then
               istat = ibr.btrv(BTRV.GETDIRECT, crcd(1, 0), crcd(0, 0), _
               iptr(4, 0), ckey(0), iptr(7, 0), cpos(0))  'get f/saved position
               istat = ibr.btrv(BTRV.GETPREV, crcd(1, 0), crcd(0, 0), _'get the
               iptr(4, 0), ckey(0), iptr(7, 0), cpos(25))   '...previous record
               i = ibr.mask(ckey(), ckey(0), ckey(20))      'mask the key value
               if istat = 0 and ckey(24) >= ckey(25) then   'valid record found
                  bc.scroll = 1                      'set scroll to scroll down
                  i = ibr.scrl(bc, ckvl(), cpos(), cvdb())'scroll the data down
                  isavrow = bc.dsprow              'save the last row displayed
                  bc.dsprow = 0                 'set row# to display at 1st row
                  i = ibr.disp(bc, cdat(), ckey(), crcd(), cvdb(), iptr())
                  bc.dsprow = isavrow + abs(isavrow < bc.rowcnt - 1)'restore it
                  lset ckvl(0) = ckey(0)          'save key value in row-buffer
                  lset cpos(0) = cpos(25)         'save the new position buffer
               else                               'a valid record was NOT found
                  ikey = KEY.HOME                       'set "user" key to HOME
                  ibypass = not 0                    'set keyboard bypass to ON
               end if
            end if
         case KEY.END                                                  'End key
            i = ibr.hkey(ckey(), ckvl(), cpos(), crcd(), iptr())  'get high key
            if cpos(0) <> cpos(24) and cpos(24) <> cpos(26) or bc.dsprow > 0 then
               lset ckey(0) = ckey(21)          'begin at highest key allowable
               istat = ibr.btrv(BTRV.GETLE, crcd(1, 0), crcd(0, 0), _
               iptr(4, 0), ckey(0), iptr(7, 0), cpos(25)) 'get record from file
               i = ibr.mask(ckey(), ckey(0), ckey(20))       'mask the file key
               if istat = 0 and ckey(24) >= ckey(25) then   'valid record found
                  bc.dsprow = 0     'set row# to 1st row and display it [below]
                  i = ibr.disp(bc, cdat(), ckey(), crcd(), cvdb(), iptr())
                  i = ibr.cdat(bc, ckey(), ckvl(), cpos(), cvdb())'clear unused
                  lset ckvl(0) = ckey(0)      'save key value in 1st-row buffer
                  lset cpos(0) = cpos(25)         'save the new position buffer
               else                               'a valid record was NOT found
                  ikey = KEY.HOME                       'set "user" key to HOME
                  ibypass = not 0                    'set keyboard bypass to ON
               end if
            else                         'already @ end or record(s) not avail.
               if cpos(24) = cpos(26) then        'data record(s) not available
                  ikey = KEY.HOME                       'set "user" key to HOME
                  ibypass = not 0                    'set keyboard bypass to ON
               end if
            end if
         case KEY.ENTER                                              'Enter key
            if iprtset then                                   'print mode is ON
               i = ibr.mask(ckey(), ckvl(0), ckey(27))       'mask the file key
               if ckey(24) >= ckey(25) then            'final key is >= 1st key
                  i = io.vmsg("Check printer for proper", bc.rowmax, bc.clrmsg)
                  i = io.vdsp("alignment and press", bc.rowmax, 25, bc.clrmsg)
                  i = io.vdsp("any key to continue.", bc.rowmax, 45, bc.clrmsg)
                  beep                                         'user-alert tone
                  if io.kget(0) <> KEY.ESC then       'OK to proceed with print
                     lset cpos(24) = cpos(0)      'save final position in range
                     i = io.vmsg("Please standby", bc.rowmax, bc.clrmsg + 128)
                     bc.linect = 99             'initialize print-lines counter
                     bc.pagect = 0              'initialize page-number counter
                     bc.prtque = not 0             'set printing active flag ON
                     lset cpos(25) = cpos(27)       'begin at 1st selected rcd.
                     istat = ibr.btrv(BTRV.GETDIRECT, crcd(1, 0), crcd(0, 0), _
                     iptr(4, 0), ckey(0), iptr(7, 0), cpos(25))      'GetRecord
                     i = ibr.mask(ckey(), ckey(0), ckvl(0))      'mask file key
                     csav = left$(ckey(24), iptr(8, 10)) 'init.key-changed var.
                     do while istat = 0 and ckey(24) <= ckey(25)'prt-range loop
                        if csav < left$(ckey(24), iptr(8, 10)) and iprtset = 1 then
                           lset csav = ckey(24)     'reset key-changed variable
                           print #bc.uniprt, ""         'print a separator line
                        end if
                        i = ibr.disp(bc, cdat(), _         'print the data-line
                        ckey(), crcd(), cvdb(), iptr())
                        if cpos(25) = cpos(24) then        'current position...
                           exit do        '...is last position; exit print-loop
                        end if
                        istat = ibr.btrv(BTRV.GETNEXT, crcd(1, 0), crcd(0, 0), _
                        iptr(4, 0), ckey(0), iptr(7, 0), cpos(25))
                        i = ibr.mask(ckey(), ckey(0), ckvl(0))        'mask key
                        if io.ktst(0) = KEY.ESC then         'abort-key pressed
                           exit do                  'abort printing - exit loop
                        end if
                     loop
                  end if
                  bc.prtque = 0               'end of print; set print mode OFF
                  iprtset = 0                     'set printing-active flag OFF
                  if bc.linect > 0 and bc.prnt = "P" then 'line(s) were printed
                     print #bc.uniprt, char(12);                'do a page-feed
                  end if
                  bc.linect = 0                      'reset print-lines counter
                  i = io.vmsg(space$(bc.colmin) + bc.cllt + _   'restore border
                  cvdb(26) + bc.clrt, bc.rowmax, bc.clrwin)     'after printing
               else                          'error: last record < first record
                  i = io.vmsg("Record selected is prior", bc.rowmax, bc.clrerr)
                  i = io.vdsp("to first record.", bc.rowmax, 25, bc.clrerr)
                  i = io.vdsp("Please retry.", bc.rowmax, 43, bc.clrerr)
                  beep                                         'user-alert tone
               end if
            else                                          'toggle or "quit" key
               if itagfld >= 0 then                         'toggle (Enter) key
                  iv = iptr(2, itagfld) - bc.padadj     'tag-field disp. length
                  iw = iptr(0, itagfld)       'pos. of tag-field in crcd(0, iy)
                  ix = iptr(1, itagfld)              'tag-field internal length
                  iy = iptr(3, itagfld) mod 100     'file# where field is found
                  do                        'begin loop to toggle & update data
                     if iy = 0 then                 'field is found in 1st file
                        lset cpos(25) = cpos(0)    'fill file-pos w/1st-row pos
                        ioper = BTRV.GETDIRECT  'set to get from saved position
                     else                         'field is found in > 1st file
                        i = ibr.fill(cdat(), ckey(), _   'fill file-key from...
                        crcd(), iptr(), ipos, iy)       '..1st-file data record
                        ioper = BTRV.GETEQ           'set to get from key value
                     end if
                     istat = ibr.btrv(ioper, crcd(1, iy), crcd(0, iy), _
                     iptr(4, iy), ckey(iy), iptr(7, iy), cpos(25))   'GetRecord
                     if istat then                    'record not found in file
                        exit do                    'exit the toggle/update loop
                     end if             'pass external (display) length [below]
                     iu = iv   'put replacement data into output buffer [below]
                     i = ibr.fmtx(mid$(crcd(0, iy), iw, ix), cvdb(27), _'get...
                     cdat(4, itagfld), cdat(5, itagfld), iu)'repl-string offset
                     mid$(crcd(0, iy), iw) = left$(cvdb(27), ix) 'update record
                     mid$(cvdb(0), itagcol) = mid$(cdat(6, itagfld), iu, iv)
                     i = io.vdsp(mid$(cdat(6, itagfld), iu, iv), _  'display...
                     bc.rowbeg, itagcol + bc.colmin, bc.clrbar)    '...new data
                     istat = ibr.btrv(BTRV.UPDATE, crcd(1, iy), crcd(0, iy), _
                     iptr(4, iy), ckey(iy), iptr(7, iy), cpos(25)) 'Update data
                  loop while istat              'exit if update accomplished OK
               end if
            end if
         case KEY.ESC                                               'Escape key
            if iprtset then                                   'print mode is ON
               iprtset = 0                        'set printing-active flag OFF
               ikey = 1                       'turn ESC key off (non-zero only)
               if bc.linect > 0 and bc.prnt = "P" then    'line(s) were printed
                  print #bc.uniprt, char(12);                   'do a page-feed
               end if
               bc.linect = 0                         'reset print-lines counter
               i = io.vmsg(space$(bc.colmin) + bc.cllt + _      'restore border
               cvdb(26) + bc.clrt, bc.rowmax, bc.clrwin)        'after printing
            end if
         case KEY.GFPOS                         'get-from-position using ckey()
            if cpos(25) = cpos(26) then 'position NOT available for current key
               ioper = BTRV.GETGE             'set operation to get a >= record
            else             'valid position is available for current key value
               ioper = BTRV.GETDIRECT    'set to get record from saved position
            end if
            istat = ibr.btrv(ioper, crcd(1, 0), crcd(0, 0), _'get record from..
            iptr(4, 0), ckey(0), iptr(7, 0), cpos(25))    '...current operation
            i = ibr.mask(ckey(), ckey(0), ckey(21))          'mask the file key
            if istat = 0 and ckey(24) <= ckey(25) then 'OK to process from here
               bc.dsprow = -1                          'begin at first data row
               do while istat = 0 and ckey(24) <= ckey(25) _
               and bc.dsprow < bc.rowcnt - 1     'get records until screen full
                  bc.dsprow = bc.dsprow + 1       'increment display row number
                  i = ibr.disp(bc, cdat(), ckey(), crcd(), cvdb(), iptr())
                  lset ckvl(bc.dsprow) = ckey(0)    'save key in current buffer
                  lset cpos(bc.dsprow) = cpos(25)  'save pos. in current buffer
                  istat = ibr.btrv(BTRV.GETNEXT, crcd(1, 0), crcd(0, 0), _
                  iptr(4, 0), ckey(0), iptr(7, 0), cpos(25))   'get next record
                  i = ibr.mask(ckey(), ckey(0), ckey(21))    'mask the file key
               loop
               if bc.dsprow < bc.rowcnt - 1 then        'EOF before screen full
                  i = ibr.cdat(bc, ckey(), ckvl(), cpos(), cvdb())
               end if                           'clear unused data rows [above]
            else                                           'record(s) not found
               ikey = KEY.END                            'set "user" key to END
               ibypass = not 0                       'set keyboard bypass to ON
            end if
         case KEY.HOME                                                'Home key
            i = ibr.lkey(ckey(), ckvl(), cpos(), crcd(), iptr())   'get low key
            if cpos(0) <> cpos(26) and (cpos(0) <> cpos(23) or inewrcd) then
               isavrow = bc.dsprow                    'save last row# displayed
               if inewrcd then                  'NEW record selection available
                  bc.dsprow = 0                     'set display row to 1st row
                  bc.dspnew = not 0          'set flag to display NEW selection
                  i = ibr.disp(bc, cdat(), ckey(), crcd(), cvdb(), iptr())
                  lset ckvl(0) = ckey(26)      'NEW selection key value = NULLs
                  lset cpos(0) = cpos(26) 'NEW selection position value = NULLs
               else                         'NEW record selection NOT available
                  bc.dsprow = -1                   'initialize first row number
               end if
               lset ckey(0) = ckey(20)     'begin at lowest allowable key value
               istat = ibr.btrv(BTRV.GETGE, crcd(1, 0), crcd(0, 0), _
               iptr(4, 0), ckey(0), iptr(7, 0), cpos(25)) 'get the first record
               i = ibr.mask(ckey(), ckey(0), ckey(21))       'mask the file key
               if istat = 0 and ckey(24) <= ckey(25) then   'valid record found
                  do while istat = 0 and ckey(24) <= ckey(25) _
                  and bc.dsprow < bc.rowcnt - 1     'get data until screen full
                     bc.dsprow = bc.dsprow + 1    'increment display row number
                     i = ibr.disp(bc, cdat(), ckey(), crcd(), cvdb(), iptr())
                     lset ckvl(bc.dsprow) = ckey(0) 'save key in current buffer
                     lset cpos(bc.dsprow) = cpos(25)     'save the new position
                     istat = ibr.btrv(BTRV.GETNEXT, crcd(1, 0), crcd(0, 0), _
                     iptr(4, 0), ckey(0), iptr(7, 0), cpos(25))'get next record
                     i = ibr.mask(ckey(), ckey(0), ckey(21)) 'mask the file key
                  loop
               else                               'a valid record was NOT found
                  if not inewrcd then       'NEW record selection not available
                     if isavrow >= 0 then              'some data WAS displayed
                        bc.dsprow = -1         'set row# to clear ALL data rows
                        i = ibr.cdat(bc, ckey(), ckvl(), cpos(), cvdb())
                     end if              'clear all data rows on screen [above]
                     if len(cdat(0, 2)) then  'disp. "No records" msg. if appl.
                        i = io.vmsg(cdat(0, 2), bc.rowmax, bc.clrerr)
                        beep                                   'user-alert tone
                        i = io.kget(0)              'wait for key after message
                     end if
                     lset ckvl(0) = ckey(26)          'return NULLs in pass-key
                     lset cpos(0) = cpos(26)   'return NULLs in position buffer
                     ikey = -1                    'no-records return key number
                     exit do                                  'exit-from-Browse
                  end if
               end if
               if bc.dsprow < isavrow and bc.auxkey <> 0 then'data rows deleted
                  i = ibr.cdat(bc, ckey(), ckvl(), cpos(), cvdb())
               end if                           'clear unused data rows [above]
            end if
         case KEY.PGDN                                           'Page-down key
            i = ibr.hkey(ckey(), ckvl(), cpos(), crcd(), iptr())  'get high key
            if cpos(0) <> cpos(24) and cpos(24) <> cpos(26) or bc.dsprow > 0 then
               if cpos(bc.dsprow) <> cpos(24) then        'add'l data available
                  istat = ibr.btrv(BTRV.GETDIRECT, crcd(1, 0), crcd(0, 0), _
                  iptr(4, 0), ckey(0), iptr(7, 0), cpos(bc.dsprow)) 'get f/posn.
                  if bc.holdln then            'hold-line-for-paging flag is ON
                     lset cpos(25) = cpos(bc.dsprow)        'set for GFPOS call
                     istat = 0         'set get-from-current oper. (do nothing)
                  else                        'hold-line-for-paging flag is OFF
                     istat = ibr.btrv(BTRV.GETNEXT, crcd(1, 0), crcd(0, 0), _
                     iptr(4, 0), ckey(0), iptr(7, 0), cpos(25)) 'get f/last row
                  end if
                  if istat then                   'next record is NOT available
                     ikey = KEY.END                      'set "user" key to END
                  else                                'next record is available
                     ikey = KEY.GFPOS                  'set "user" key to GFPOS
                  end if
               else                                   'add'l data NOT available
                  ikey = KEY.END                         'set "user" key to END
               end if
               ibypass = not 0                       'set keyboard bypass to ON
            else                        'add'l data NOT available for page-down
               if cpos(24) = cpos(26) then          'no record(s) are available
                  ikey = KEY.HOME                       'set "user" key to HOME
                  ibypass = not 0                    'set keyboard bypass to ON
               end if
            end if
         case KEY.PGUP                                             'Page-up key
            i = ibr.lkey(ckey(), ckvl(), cpos(), crcd(), iptr())   'get low key
            if cpos(0) <> cpos(26) and (cpos(0) <> cpos(23) or inewrcd) then
               ikeycnt = 0                         'initialize file-key counter
               istat = ibr.btrv(BTRV.GETDIRECT, crcd(1, 0), crcd(0, 0), _
               iptr(4, 0), ckey(0), iptr(7, 0), cpos(0))  'get f/saved position
               i = ibr.mask(ckey(), ckey(0), ckey(20))       'mask the file key
               if istat = 0 and ckey(24) >= ckey(25) then   'prior record found
                  do while istat = 0 and ckey(24) >= ckey(25) _
                  and ikeycnt < bc.rowcnt - abs(bc.holdln) 'get pageful of data
                     ikeycnt = ikeycnt + 1          'increment file-key counter
                     istat = ibr.btrv(BTRV.GETPREV, crcd(1, 0), crcd(0, 0), _
                     iptr(4, 0), ckey(0), iptr(7, 0), cpos(25))'get prev.record
                     i = ibr.mask(ckey(), ckey(0), ckey(20)) 'mask the file key
                  loop
                  if ikeycnt >= bc.rowcnt - abs(bc.holdln) or not inewrcd then
                     ikey = KEY.GFPOS                  'set "user" key to GFPOS
                  else                 'less than full page of prior data found
                     ikey = KEY.HOME                    'set "user" key to HOME
                  end if
               else                                     'prior record NOT found
                  ikey = KEY.HOME                       'set "user" key to HOME
               end if
               ibypass = not 0                       'set keyboard bypass to ON
            end if
         case KEY.PRINT, _                           'Print-activation key (F9)
              KEY.PRINT2                            'Print-activation key (F10)
            if bc.prinit = 0 then                 'printing not yet initialized
               i = io.vmsg("Print to (P)rinter or (D)isk?", bc.rowmax, bc.clrmsg)
               do                                  'loop until selection or Esc
                  beep                                         'user-alert tone
                  lset bc.prnt = ucase$(char(io.kget(0)))    'wait for user key
                  if asc(bc.prnt) = KEY.ESC then  'print-dest. selection or ESC
                     exit do                      'user pressed ESC - exit loop
                  end if
                  if bc.prnt = "P" or bc.prnt = "D" then 'dest.=printer or disk
                     bc.uniprt = freefile                   'save file-unit no.
                     if bc.prnt = "P" then                'printer was selected
                        open "lpt1" for output as bc.uniprt   'open the printer
                     else                                 'disk output selected
                        open "browse.prt" for output as bc.uniprt    'open file
                     end if
                     bc.prinit = ikey - KEY.PRINT + 1  'set print-init. flag ON
                     exit do               'exit the destination-selection loop
                  end if
               loop
            end if
            if bc.prinit then                    'printing has been initialized
               iprtset = bc.prinit                   'set print-mode flag to ON
               lset ckey(27) = ckvl(0)              'save the current key value
               lset cpos(27) = cpos(0)               'save the current position
               i = io.vmsg("Move bar to last record to print", bc.rowmax, bc.clrmsg)
               i = io.vdsp("and press Enter to select.", bc.rowmax, 33, bc.clrmsg)
               i = io.vdsp("Esc-exit print.", bc.rowmax, 61, bc.clrmsg)
               beep                                            'user-alert tone
            else
               i = io.vmsg(space$(bc.colmin) + bc.cllt + _      'restore border
               cvdb(26) + bc.clrt, bc.rowmax, bc.clrwin)      'when Esc-key hit
            end if
         case else       'alpha (search) keys; print-begin or toggle; exit keys
            if ikey >= 32 and ikey <= 127 then             'alpha (search) keys
               if (iedtfld < bc.edtmax or iedtcol < iptr(6, bc.edtmax + 10)) _
               and bc.edtmax >= 0 then            'OK to process the search key
                  icmp = not 0                  'initialize buffer-compare flag
                  ibufkey = not 0                'set buffered keyboard flag ON
                  if iedtcol < iptr(6, iedtfld + 10) then  'edit position is OK
                     mid$(cdat(7, iedtfld), iedtcol - iptr(5, iedtfld + 10) _
                     + 1) = char(ikey)     'put search character to edit buffer
                  else                'edit position is at end of current field
                     mid$(cdat(7, iedtfld + 1), 1) = char(ikey) 'use next field
                  end if    'edit position is 1st position in 1st field [below]
                  if iedtcol = iptr(5, 10) then'turn on 1st-fld.display [below]
                     i = io.vdsp(cdat(7, 0), iedtrow, iedtcol, bc.clredt)
                  else          'edit position is NOT 1st position in 1st field
                     i = io.vdsp(char(ikey), iedtrow, iedtcol, bc.clredt)
                  end if            'display just the current character [above]
                  iedtcol = iedtcol + 1    'increment the current edit position
                  if iedtcol = iptr(6, iedtfld + 10) and iedtfld < bc.edtmax then
                     iedtfld = iedtfld + 1'edit pos.@field end;increment field#
                     iedtcol = iptr(5, iedtfld + 10)'edit @1st pos.in new field
                     i = io.vdsp(cdat(7, iedtfld), iedtrow, iedtcol, bc.clredt)
                  end if             'turn on display of new edit field [above]
                  iedtlen = iedtcol - iptr(5, iedtfld + 10)'current edit length
                  if iedtlen > 0 and iedtlen < iptr(6, iedtfld) and _
                  instr("D5T5N4N2", cdat(4, iptr(5, iedtfld))) then
                     mid$(cdat(7, iedtfld), iedtlen + 1) = "000000000"
                  end if      'numeric-field data entered; do zero-fill [above]
                  lset ckey(0) = cdat(7, 10)        'put left-pad data into key
                  ipos = len(cdat(7, 10)) + 1     'initialize key pointer w/pad
                  iend = iedtfld - abs(iedtlen = 0) 'last edit-field to process
                  if iedtlen = 0 and iend >= 0 then   'edit-pos. at field-begin
                     iedtlen = iptr(6, iend)   'set edit-length to field length
                  end if
                  for indx = 0 to iend                 'loop thru search fields
                     ifld = iptr(5, indx)    'temp. copy of search-field number
                     if indx = iend then               'last edit field in loop
                        if cdat(4, ifld) = "E1" then 'search-field type is "E1"
                           csep = cdat(5, ifld) + cdat(6, ifld)'sep+segm.length
                        else            'search-field types are other than "E1"
                           csep = bc.csep             'separator character only
                        end if                   'format the edit-field [below]
                        if cdat(4, ifld) = "D1" then 'search-field type is "D1"
                           i = ibr.fmtx(mid$(ckvl(0), ipos + 2, 4) + _
                                        mid$(ckvl(0), ipos + 0, 2), _
                                   cvdb(27), cdat(4, ifld), csep, iptr(6, indx))
                        else            'search-field types are other than "D1"
                           i = ibr.fmtx(mid$(ckvl(0), ipos, iptr(1, ifld)), _
                           cvdb(27), cdat(4, ifld), csep, iptr(6, indx))
                        end if
                        if midchar(ckey(10), ipos) = 67 then 'uppercase masking
                           if ucase$(left$(cdat(7, indx), iedtlen)) = _
                           ucase$(left$(cvdb(27), iedtlen)) then
                              icmp = 0         'data match; set search flag OFF
                              exit for    'exit the key-match search-field loop
                           end if
                        else                 'not an uppercase-mask ("C") field
                           if left$(cdat(7, indx), iedtlen) = _
                           left$(cvdb(27), iedtlen) then
                              icmp = 0         'data match; set search flag OFF
                              exit for    'exit the key-match search-field loop
                           end if
                        end if
                     end if
                     i = ibr.fmti(cdat(7, indx), cvdb(27), cdat(4, ifld), _
                     cdat(6, ifld))           'reformat data to internal format
                     mid$(ckey(0), ipos) = left$(cvdb(27), iptr(1, ifld))
                     ipos = ipos + iptr(8, indx + 10)    'increment key-pointer
                  next
                  if icmp then                          'file-search flag is ON
                     if ipos <= ikeylen then          'key-pointer < key-length
                        mid$(ckey(0), ipos) = ckey(26) 'fill out key with NULLs
                     end if
                     istat = ibr.btrv(BTRV.GETGE, crcd(1, 0), crcd(0, 0), _
                     iptr(4, 0), ckey(0), iptr(7, 0), cpos(25))  'get file data
                     if istat then           'status = EOF (no further records)
                        ikey = KEY.END                   'set "user" key to END
                        ibypass = not 0              'set keyboard bypass to ON
                     else                           'subsequent record(s) found
                        i = ibr.mask(ckey(), ckey(0), ckey(20))  'mask file key
                        if ckey(0) <> ckvl(0) then      'data found <> 1st line
                           if ckey(24) < ckey(25) then 'data found < data range
                              ikey = KEY.HOME           'set "user" key to HOME
                           else                   'data found within data range
                              ikey = KEY.GFPOS         'set "user" key to GFPOS
                           end if
                           ibypass = not 0           'set keyboard bypass to ON
                        end if
                     end if
                  end if
               end if             'end of "OK-to-process-search-key" selections
            end if                        'end of alpha (search-key) selections
      end select               'end of user-hit or default key-value selections
   loop  while bc.auxkey <> 0 _            'Browse "enter, display, & exit" key
   and ikey <> bc.xkey01 _                                'Browse "quit" key #1
   and ikey <> bc.xkey02 _                                'Browse "quit" key #2
   and ikey <> bc.xkey03                                  'Browse "quit" key #3
   bc.auxkey = ikey                             'return key# to calling program
   lset ckey(0) = ckvl(0)                'return 1st-row key to calling program
   for indx = 0 to bc.edtmax                       'loop thru the search-fields
      iptr(5, indx + 10) = iptr(5, indx + 10) - bc.colmin  'reset begin-columns
   next
   for indx = bc.fldmin to bc.fldmax              'loop thru the display-fields
      iptr(2, indx) = iptr(2, indx) - bc.padadj         'reset the field-widths
   next
end function

function ibr.btrv(ioper, cfcb, cbuf, ilen, ckey, ikey, cpos) 'perform BTRV call
   if ioper = BTRV.GETDIRECT then    'get record from previously saved position
      lset cbuf = cpos                   'put saved position into record buffer
   else                                               'standard "get" operation
      lset cbuf = ""                       'pre-clear the Btrieve record buffer
   end if
   istat = BTRV(ioper, cfcb, cbuf, (ilen), ckey, ikey)  'call Btrieve rcd. mgr.
   if ioper <> BTRV.GETDIRECT and ioper <> BTRV.UPDATE _
   and ioper < 50 and istat = 0 then    'OK to get the physical record position
      istat = BTRV(BTRV.GETPOS, cfcb, cpos, 4, ckey, ikey) '..physical position
   end if                    'unrecoverable Btrieve error [below]; exit program
   if istat <> 0 and istat <> BTRV.EOF and istat <> BTRV.NOTFND then
      print
      print "Btrieve opcode = "; ltrim$(str$(ioper))    'display operation code
      print "Btrieve error  = "; ltrim$(str$(istat))      'display error number
      beep                                               'sound User-alert tone
      i = BTRV(BTRV.RESET, "", "", 0, "", 0)        'call BTRV to perform reset
      close                                                    'close all files
      system                                                     'return to DOS
   end if
   ibr.btrv = istat                     'return BTRV status to calling function
end function

function ibr.cdat(bc as bcnf, ckey(), ckvl(), cpos(), cvdb())'clear unused data
   for irow = bc.dsprow + 1 to bc.rowcnt - 1        'loop thru unused data-rows
      lset ckvl(irow) = ckey(26)                     'clear data-row key buffer
      lset cpos(irow) = cpos(26)                'clear data-row position buffer
      lset cvdb(irow) = cvdb(23)                   'clear data-row video buffer
      i = io.vdsp(cvdb(23), bc.rowbeg + irow, bc.colmin + 1, bc.clrdat)
   next
end function

function ibr.disp(bc as bcnf, cdat(), ckey(), crcd(), cvdb(), iptr())'disp.data
   ipadadj = abs(bc.padint) * 2   'used to right-justify printed numeric fields
   ibegfld = bc.fldmin             'begin formatting with first printable field
   if bc.prtque then                            'print-mode ON; printing active
      ifldcol = 1                          'first position in print-line buffer
   else                             'display-mode formatting; printing inactive
      ifldcol = 1 + abs(bc.padext) + abs(bc.padint)'1st pos.in disp-line buffer
   end if
   if bc.prtque and bc.linect >= 59 then        'print-mode ON; printing active
      bc.pagect = bc.pagect + 1                         'increment page-counter
      lset bc.page = ltrim$(str$(bc.pagect))          'fill page-counter buffer
      lset bc.time = time$                     'put current time to time buffer
      lset cprt = date$                       'put current date to print buffer
      mid$(cprt, bc.hdrpos) = cdat(0, 0)      'put header title to print buffer
      mid$(cprt, 70) = "Page: "           'put page-field title to print buffer
      mid$(cprt, 76) = bc.page                 'put page number to print buffer
      if bc.pagect > 1 then                               'current page# is > 1
         print #bc.uniprt, char(12);                  'page-feed when page# > 1
      end if
      print #bc.uniprt, cprt              'print date, title, page# on 1st line
      print #bc.uniprt, bc.time                     'print time-of-day 2nd line
      print #bc.uniprt, ""                           'print blank line 3rd line
      if bc.titles then                      'titles available; print title row
         lset cvdb(25) = ""                             'clear the print buffer
         ipos = ifldcol                               'beginning print position
         for ifld = ibegfld to bc.fldmax   'loop thru the data fields (columns)
            if instr("AHIMN", left$(cdat(4, ifld), 1)) then      'numeric field
               itab = iptr(2, ifld) - len(cdat(2, ifld)) - ipadadj  'tab column
               if itab < 0 then           'title is same length as padded field
                  itab = 0               'don't tab over from field left column
               end if
            else                                             'non-numeric field
               itab = 0                                          'tab-to-column
            end if
            mid$(cvdb(25), ipos + itab) = cdat(2, ifld)        'print the title
            ipos = ipos + iptr(2, ifld) + 1        'move to next print position
         next                           'increment to next start-column [above]
         print #bc.uniprt, rtrim$(cvdb(25))           'terminate the title-line
         if bc.titles = 3 then                         '2nd title-row available
            lset cvdb(25) = ""                          'clear the print buffer
            ipos = 1                                  'beginning print position
            for ifld = ibegfld to bc.fldmax    'loop thru data fields (columns)
               if instr("AHIMN", left$(cdat(4, ifld), 1)) then   'numeric field
                  itab = iptr(2, ifld) - len(cdat(3, ifld)) - ipadadj'tab column
                  if itab < 0 then        'title is same length as padded field
                     itab = 0            'don't tab over from field left column
                  end if
               else                                          'non-numeric field
                  itab = 0                                       'tab-to-column
               end if
               mid$(cvdb(25), ipos + itab) = cdat(3, ifld)     'print the title
               ipos = ipos + iptr(2, ifld) + 1     'move to next print position
            next
            print #bc.uniprt, rtrim$(cvdb(25))    'terminate the 2nd title-line
         end if
         lset cvdb(25) = ""                             'clear the print buffer
         ipos = 1                                     'beginning print position
         for ifld = ibegfld to bc.fldmax       'loop thru data fields (columns)
            mid$(cvdb(25), ipos) = string$(iptr(2, ifld), "-") 'print underline
            ipos = ipos + iptr(2, ifld) + 1        'move to next print position
         next                            'put dashed lines under titles [above]
         print #bc.uniprt, rtrim$(cvdb(25))    'terminate the titles-underscore
         lset cvdb(25) = ""                             'clear the print buffer
      end if
      bc.linect = 3 + bc.titles                    'accumulated lines this page
   end if
   ivarfld = -99                           'init.variable-length text field no.
   for ifld = ibegfld to bc.fldmax         'loop thru the data fields (columns)
      if bc.dspnew then                            'OK to display NEW selection
         if ifld = 0 then                          'field-number is first field
            lset cdat(1, ifld) = cdat(0, 1) 'put NEW msg. to video field-buffer
         else                                        'field-number == 1st field
            lset cdat(1, ifld) = ""                   'clear video field-buffer
         end if
      else                                           'display a normal data-row
         iy = iptr(3, ifld)                'field-ptr. to selected Btrieve file
         if iy > 0 then                             'Btrieve file is > 1st file
            i = ibr.fill(cdat(), ckey(), crcd(), iptr(), ipos, iy)'fill filekey
            lset ckey(28) = ckey(iy)           'make copy of key for comparison
            i = ibr.uprc(ckey(28), ckey(iy + 10))  'uppercase-mask the key copy
            istat = ibr.btrv(BTRV.GETGE, crcd(1, iy), crcd(0, iy), _  'get a...
            iptr(4, iy), ckey(iy), iptr(7, iy), space$(4))'...>= Btrieve record
            i = ibr.uprc(ckey(iy), ckey(iy + 10))  'uppercase-mask the file key
            if istat <> 0 or left$(ckey(iy), ipos - 1) _
            <> left$(ckey(28), ipos - 1) then     'record not found or no match
               lset crcd(0, iy) = ""           'clear the related-record buffer
            end if
         end if
         ifldtpc = midchar(cdat(4, ifld), 1)    'Hn':ifldtpc=72;'Mn':ifldtpc=77
         ifldtpn = midchar(cdat(4, ifld), 2) - 48     'if type='S1', ifldtpn =1
         iextlen = iptr(2, ifld) - bc.padadj      'data-length in display-field
         iintlen = iptr(1, ifld)          'internal field-length in data record
         ioffset = iptr(0, ifld)            'offset = begin-byte in data-record
         select case cdat(4, ifld)                         'select on data-type
            case "A1"                                    'ASCII byte (0 to 255)
               rset cdat(1, ifld) = right$(str$(midchar(crcd(0, iy), ioffset)), _
               iextlen)                       'put data into video field-buffer
            case "C1", _                          'character (alpha) type field
                 "C2"                         'character type (variable length)
               if cdat(4, ifld) = "C2" then   'character type (variable length)
                  ivarfld = ifld            'set variable-length text field no.
               end if
               lset cdat(1, ifld) = mid$(crcd(0, iy), ioffset, iintlen)
               for ix = 1 to len(cdat(1, ifld))   'loop thru video field-buffer
                  if midchar(cdat(1, ifld), ix) < 32 then 'control' character..
                     mid$(cdat(1, ifld), ix) = "."  '...found in Btrieve buffer
                  end if          'substitute '.' for control character [above]
               next
            case "D1", _                          'date type mmddyy or mmddyyyy
                 "D2", _                          'date type yymmdd or yyyymmdd
                 "D5", _                              'date type dmyy (Btrieve)
                 "E1", _                      'expandable char.(phone, zip,...)
                 "T1", _                              'time type hhmm or hhmmss
                 "T5"                                 'time type ssmh (Btrieve)
               if cdat(4, ifld) = "E1" then         'display-field type is "E1"
                  csep = cdat(5, ifld) + cdat(6, ifld)  'separator+segm.lengths
               else                      'display-field type is other than "E1"
                  csep = bc.csep                      'separator character only
               end if                         'format the display-field [below]
               i = ibr.fmtx(mid$(crcd(0, iy), ioffset, iintlen), _
               cdat(1, ifld), cdat(4, ifld), csep, iextlen)
            case "H6", _                      'Double IEEE extendable [**] type
                 "H7", _                      'Double IEEE extendable [*/] type
                 "H8", _                      'Double IEEE extendable [/*] type
                 "H9", _                      'Double IEEE extendable [//] type
                 "I5", _                        'Single IEEE float numeric type
                 "I6", _                           'same as I5 plus zero-filled
                 "I7", _                        'Double IEEE float numeric type
                 "I8", _                           'same as I7 plus zero-filled
                 "M6", _                       'Double MBF extendable [**] type
                 "M7", _                       'Double MBF extendable [*/] type
                 "M8", _                       'Double MBF extendable [/*] type
                 "M9", _                       'Double MBF extendable [//] type
                 "N0", _                        'Double MBF dollar & cents type
                 "N1", _                            'Short integer numeric type
                 "N2", _                           'same as N1 plus zero-filled
                 "N3", _                             'Long integer numeric type
                 "N4", _                           'same as N3 plus zero-filled
                 "N5", _                               'Single MBF numeric type
                 "N6", _                           'same as N5 plus zero-filled
                 "N7", _                               'Double MBF numeric type
                 "N8"                              'same as N7 plus zero-filled
               select case ifldtpn                       'reselect on data-type
                  case 1, _               'integer-type field without zero-fill
                       2                     'integer-type field with zero-fill
                     rset cdat(1, ifld) = right$(str$(cvi(mid$(crcd(0, iy), _
                     ioffset, 2))), iextlen)  'put data into video field-buffer
                  case 3, _          'long-integer-type field without zero-fill
                       4                'long-integer-type field with zero-fill
                     rset cdat(1, ifld) = right$(str$(cvl(mid$(crcd(0, iy), _
                     ioffset, 4))), iextlen)  'put data into video field-buffer
                  case else                   'case 0, 5 - 8; float-type fields
                     if ifldtpn = 0 then         'dollar/cents float-type field
                        idec = 2           'set decimal places for dollar/cents
                     else                   'non-dollar/cents float-type fields
                        idec = bc.declen  '#places specified by calling program
                     end if                   'single-float type fields [below]
                     if (ifldtpn = 5 or ifldtpn = 6) and _
                     (not(ifldtpc = 72 or ifldtpc = 77)) then    'not 'Hn'/'Mn'
                        if ifldtpc = 73 then                    'I5','I6' types
                           dval = cvs(mid$(crcd(0, iy), ioffset, 4))'IEEE->num.
                        else                                    'N5','N6' types
                           dval = fcvs(mid$(crcd(0, iy), ioffset, 4))'MBF->num.
                        end if
                     else                        'double-precision float fields
                        if ifldtpc = 72 or ifldtpc = 73 then    'Hn','In' types
                           dval = cvd(mid$(crcd(0, iy), ioffset, 8))'IEEE->num.
                        else                          'Mn','N0','N7','N8' types
                           dval = dcvd(mid$(crcd(0, iy), ioffset, 8))'MBF->num.
                        end if
                        if ifldtpc = 72 or ifldtpc = 77 then    'Hn','Mn' types
                           ival = pdqvali(cdat(5, ifld)) + bc.fldmin'ext.field#
                           if ifldtpn <= 7 then     'extendable [*] type fields
                              if ival >= bc.fldmin then 'valid ext. field found
                                 dval = val(cdat(1, ival)) * dval
                              end if  'multiply prev.field by extension [above]
                           else                     'extendable [/] type fields
                              if ival >= bc.fldmin then 'valid ext. field found
                                 if dval then         'non-zero extension found
                                    dval = val(cdat(1, ival)) / dval
                                 end if 'divide prev.field by extension [above]
                              end if
                           end if
                           if ifldtpn mod 2 then                 ' M7, M9 types
                              dval = dval / val(cdat(6, ifld))'div. by constant
                           else                                  ' M6, M8 types
                              dval = dval * val(cdat(6, ifld))'mult.by constant
                           end if
                        end if
                     end if      'pre-add rounding to float-type fields [below]
                     dval = dval + val("." + string$(idec, "0") + "5") * sgn(dval)
                     lset cvdb(27) = ltrim$(str$(dval))    'put to video buffer
                     ichr = instr(cvdb(27), char(68))  'exponent char. position
                     if ichr then   'exponent char.found; get magnitude [below]
                        izfl = pdqvali(mid$(cvdb(27), ichr + 2)) - 1
                        lset cvdb(27) = "0." + string$(izfl, "0") + _
                        left$(cvdb(27), 1) + mid$(cvdb(27), 3, ichr - 3)
                     else             'exponent char. NOT found in video buffer
                        if asc(cvdb(27)) = 46 then   'video buffer begins w/"."
                           lset cvdb(27) = "0" + cvdb(27) 'insert "0" in buffer
                        end if
                     end if   'replace exponent format w/decimal format [above]
                     ilen = len(rtrim$(cvdb(27)))     'trimmed length of buffer
                     idot = instr(cvdb(27), ".")     'position of decimal point
                     if idot = 0 then        'decimal point NOT found in buffer
                        idot = ilen + 1          'set position to end of buffer
                        izfl = idec - ilen + idot - 1    '# zeros to pad format
                     else              'a decimal point was found in the buffer
                        izfl = idec - ilen + idot    '# zeros to pad the format
                     end if
                     if izfl < 0 then  'decimal places already exceed bc.declen
                        izfl = 0     'set no. of zeros to add to buffer to zero
                     end if          'format and right-justify the data [below]
                     rset cdat(1, ifld) = left$(cvdb(27), idot - 1) + "." _
                                  + mid$(rtrim$(cvdb(27)), idot + 1, idec) _
                                  + string$(izfl, "0")
               end select                        'zero-fill field-types [below]
               if ifldtpn > 0 and ifldtpn mod 2 = 0 and _
               (not(ifldtpc = 72 or ifldtpc = 77)) then          'not 'Hn'/'Mn'
                  mid$(cdat(1, ifld), 1) = string$(iextlen - _
                  len(ltrim$(cdat(1, ifld))), "0") 'zero-fill the display field
               end if
            case "N9"              'right-justified numeric stored as character
               rset cdat(1, ifld) = right$(mid$(crcd(0, iy), _
               ioffset, iintlen), iextlen)      'put data to video field-buffer
            case "S0", _                 'replaceable field-type (index-base 0)
                 "S1", _                 'replaceable field-type (index-base 1)
                 "S2", _                  'updateable field-type (index-base 0)
                 "S3", _                  'updateable field-type (index-base 1)
                 "S4", _                 'updateable field-type (source lookup)
                 "S5"                   'replaceable field-type (source lookup)
               if ioffset = 0 then        'a begin-byte# has NOT been specified
                  iintlen = 0       'set length to pass NULL source to function
                  ioffset = 1        'set offset to avoid invalid function call
               end if
               iselptr = iextlen    'pass external (display) length to function
               i = ibr.fmtx(mid$(crcd(0, iy), ioffset, iintlen), "", _
               cdat(4, ifld), cdat(5, ifld), iselptr)  'get repl.-string offset
               lset cdat(1, ifld) = mid$(cdat(6, ifld), iselptr, iextlen)
            case "S6"                         'replaceable type (const. subst.)
               mid$(crcd(0, iy), ioffset) = cdat(6, ifld)'put subst.into record
               lset cdat(1, ifld) = cdat(6, ifld) 'put subst. into field buffer
            case else
         end select
      end if
      if bc.prtque then                         'print-mode ON; printing active
         if cdat(4, ifld) <> "C2" then       ' don't print var.length data here
            mid$(cvdb(25), ifldcol) = cdat(1, ifld)'put field-data print-buffer
            ifldcol = ifldcol + iptr(2, ifld) + 1 'incr.disp./print buffer posn.
         end if
      elseif ifld >= 0 then                    'print mode is OFF, display only
         mid$(cvdb(24), ifldcol) = cdat(1, ifld)'put field-data to video-buffer
         ifldcol = ifldcol + iptr(2, ifld) + 1'incr.disp./print buffer position
      end if
   next
   if bc.prtque then                            'print-mode ON; printing active
      print #bc.uniprt, rtrim$(cvdb(25))              'print the main data line
      if bc.prnt = "P" then                          'printer output designated
         bc.linect = bc.linect + 1               'increment print-lines counter
      end if
      if ivarfld > -99 then                    'variable length character field
         ilen = len(rtrim$(crcd(0, 0)))   'actual variable-length buffer length
         iy = iptr(0, ivarfld)            'set text begin to data record offset
         if iy + iptr(1, ivarfld) <= ilen then 'text was NOT completely printed
            for ix = iy to ilen              'loop thru the Btrieve data buffer
               if midchar(crcd(0, 0), ix) < 32 then      'control' character...
                  mid$(crcd(0, 0), ix) = " "        '...found in Btrieve buffer
               end if             'substitute ' ' for control character [above]
            next
            ilen = len(rtrim$(crcd(0, 0)))'actual variable-length buffer length
            while iy <= ilen              'loop to print the text line segments
               ix = iy                    'print begin == last char.printed + 1
               iy = iy + 80               'initialize the ending print position
               if iy <= ilen then         'ending print position is within text
                  while iy > ix and midchar(crcd(0, 0), iy) <> 32'find ' ' pos.
                     iy = iy - 1           'decrement the ending print position
                  wend          'loop until ' ' character found or end == begin
               end if
               if iy = ix then      'end print position == begin print position
                  iy = iy + 80        'reset the ending print position as above
               end if
               if iy > ilen then        'end pos.beyond last pos.in data buffer
                  iy = ilen + 1         'set end pos.to last pos.in data buffer
               end if            'see if text line segment is non-blank [below]
               if ltrim$(mid$(crcd(0, 0), ix, iy - ix)) <> "" then  'non-blank!
                  print #bc.uniprt, "    "; mid$(crcd(0, 0), ix, iy - ix)
               end if                     'output the text line segment [above]
            wend
         end if
      end if
   else                                        'print mode is OFF, display only
      bc.dspnew = 0                              'reset display-NEW-record flag
      lset cvdb(bc.dsprow) = cvdb(24)           'put display-data to row-buffer
      if bc.dsprow = 0 and bc.auxkey <> 0 then 'bc.auxkey set to stay in Browse
         icolor = bc.clrbar                  'set row-color to scroll-bar color
      else                                     'bc.auxkey set to display & exit
         icolor = bc.clrdat                        'set row-color to data color
      end if                                      'display the data-row [below]
      i = io.vdsp(cvdb(24), bc.rowbeg + bc.dsprow, bc.colmin + 1, icolor)
   end if
end function

function ibr.fill(cdat(), ckey(), crcd(), iptr(), ipos, iy)  'fill key f/record
   ix = iy \ 100 + (iy >= 100)         'file number from which to load key data
   iy = iy mod 100                    'remove file number padding if applicable
   mid$(ckey(iy), 1) = cdat(7, iy + 10) + string$(len(ckey(iy)), 0)   'left-pad
   ipos = len(cdat(7, iy + 10)) + 1             'begin pos. after left-pad data
   for iz = 0 to iptr(7, iy + 10)                   'loop thru the key segments
      ival1 = iptr(iy + 8, iz)              'seg. begin pos. in 1st-file record
      ival2 = iptr(iy + 8, iz + 10)          'segment length in 1st-file record
      if ival1 < 0 then                       'format mismatch; convert segment
         if midchar(ckey(iy + 10), ipos) = 73 then  'file #iy segment is 16-bit
            i = ibr.fmti(mid$(crcd(0, ix), -ival1, ival2), ckey(28), "N2", "")
            ival2 = 2              'set ival2 to increment key-segment position
         else                                    'data-record segment is 16-bit
            i = ibr.fmtx(mid$(crcd(0, ix), -ival1, 2), ckey(28), "N2", "", ival2)
         end if
         mid$(ckey(iy), ipos) = left$(ckey(28), ival2)'put value to file iy key
      else                                'key-segment and record formats match
         mid$(ckey(iy), ipos) = mid$(crcd(0, ix), ival1, ival2)'put data to key
      end if
      ipos = ipos + ival2                       'increment key-segment position
   next
end function

function ibr.fmti(csrc, cdst, ctyp, csep)     'reformat data to internal format
   select case ctyp                                         'browse field types
      case "D1", _                                'date type mmddyy or mmddyyyy
           "D2"                                   'date type yymmdd or yyyymmdd
         lset cdst = mid$(csrc, 3, 2) + mid$(csrc, 6, 2) + mid$(csrc, 9, 2)
      case "D5"                                       'date type dmyy (Btrieve)
         lset cdst = char(pdqvali(mid$(csrc, 9, 2))) + _
                     char(pdqvali(mid$(csrc, 6, 2))) + _
                     mki$(pdqvali(mid$(csrc, 1, 4)))'put data into dest.-buffer
      case "E1"                               'expandable char.(phone, zip,...)
         ipos = 1                              'begin position in source buffer
         for iseg = 1 to len(csep) + 1         'loop thru the no. of separators
            if iseg > len(csep) then        'final segment in the source buffer
               isiz = len(csrc) - ipos + 1   'datalen following final separator
            else                    'NOT the final segment in the source buffer
               isiz = midchar(csep, iseg) 'datalen to the left of the separator
            end if                   'add segment data to output buffer [below]
            mid$(cdst, ipos - iseg + 1) = mid$(csrc, ipos, isiz)
            ipos = ipos + isiz + 1   'increment source position to next segment
         next
      case "N2"                                     '16-bit integer-type fields
         if pdqvall&(csrc) > 32767 then                       'integer overflow
            ival = 32767                                  'set to maximum value
         else                                          'data entered <= maximum
            ival = pdqvali(csrc)                          'set to value entered
         end if
         lset cdst = mki$(ival)                     'put data into dest.-buffer
      case "N4"                                     '32-bit integer-type fields
         lset cdst = mkl$(pdqvall&(csrc))           'put data into dest.-buffer
      case "T5"                                       'time type ssmh (Btrieve)
         lset cdst = char(0) + _                    'put data into dest.-buffer
                     char(pdqvali(mid$(csrc, 7, 2))) + _
                     char(pdqvali(mid$(csrc, 4, 2))) + _
                     char(pdqvali(mid$(csrc, 1, 2)))
      case else                                          'character-type fields
         lset cdst = csrc                           'put data into dest.-buffer
   end select
end function

function ibr.fmtx(csrc, cdst, ctyp, csep, ilen) 'format data to external format
   select case ctyp                                         'browse field types
      case "D1", _                                'date type mmddyy or mmddyyyy
           "D2"                                   'date type yymmdd or yyyymmdd
         if ctyp = "D1" then                      'date type mmddyy or mmddyyyy
            lset csrc = mid$(csrc, 5) + left$(csrc, 4)
         end if                      'move year to front of date buffer [above]
         if len(csrc) = 6 and asc(csrc) <> 32 then  'date type mmddyy or yymmdd
            if left$(csrc, 2) >= "80" then                       'year is >= 80
               csrc = "19" + csrc                            'make year >= 1980
            else                                                 'year is <= 79
               csrc = "20" + csrc                            'make year <= 2079
            end if
         end if
         lset cdst = mid$(csrc, 1, 4) + csep + mid$(csrc, 5, 2) + csep + _
                     mid$(csrc, 7, 2)               'put data into dest.-buffer
      case "D5"                                       'date type dmyy (Btrieve)
         if cvl(csrc) = 0 or rtrim$(csrc) = "" then        'source date is NULL
            lset cdst = "    " + csep + "  " + csep   'set formatted date blank
         else                                          'source date is not NULL
            lset cdst = ltrim$(str$(cvi(mid$(csrc, 3)))) + csep + _
                        right$("0" + ltrim$(str$(midchar(csrc, 2))), 2) + csep + _
                        right$("0" + ltrim$(str$(midchar(csrc, 1))), 2)
         end if                             'put data into dest.-buffer [above]
      case "E1"                               'expandable char.(phone, zip,...)
         ipos = 1                         'begin position in source data record
         isep = len(csep) \ 2             'total length of separator characters
         for iseg = 1 to isep + 1              'loop thru the no. of separators
            if iseg > isep then             'final segment in the source buffer
               isiz = len(csrc) - ipos + 1   'datalen following final separator
            else                    'NOT the final segment in the source buffer
               isiz = midchar(csep, iseg + isep)  'datalen to left of separator
               mid$(cdst, ipos + iseg + isiz - 1) = char(midchar(csep, iseg))
            end if              'add segment separator to output buffer [above]
            mid$(cdst, ipos + iseg - 1) = mid$(csrc, ipos, isiz)'add segm. data
            ipos = ipos + isiz       'increment source position to next segment
         next
      case "N2"                                     '16-bit integer-type fields
         lset cdst = right$("0000" + ltrim$(str$(cvi(csrc))), ilen)
      case "N4"                                     '32-bit integer-type fields
         lset cdst = right$("000000000" + ltrim$(str$(cvl(csrc))), ilen)
      case "S0", _                       'replaceable field-type (index-base 0)
           "S1", _                       'replaceable field-type (index-base 1)
           "S2", _                        'updateable field-type (index-base 0)
           "S3", _                        'updateable field-type (index-base 1)
           "S4", _                       'updateable field-type (source lookup)
           "S5"                         'replaceable field-type (source lookup)
         isep = len(csep)                       'source selection-string length
         isrc = len(csrc)                        'internal (source) data length
         ityp = midchar(ctyp, 2)            'index-base specifier in field type
         if ityp > 3 then                            'source-lookup-type fields
            for ipos = 1 to isep step isrc          'scan the source selections
               if mid$(csep, ipos, isrc) = csrc then'select-data matches source
                  exit for                        'exit the selection-scan loop
               end if
            next
         else                                           'index-base-type fields
            select case isrc               'valid source data has been provided
               case 0                  'valid source data has NOT been provided
                  ipos = 1       'set position to select fixed replacement data
                  isrc = 1                 'reset internal (source) data length
               case 1                               'source is 8-bit ASCII data
                  ipos = (asc(csrc) - ityp mod 2) * isrc + 1 'repl.-string pos.
               case 2                            'source is 16-bit integer data
                  ipos = (cvi(csrc) - ityp mod 2) * isrc + 1 'repl.-string pos.
               case else
            end select
         end if
         if ipos > 0 then                 'replacement-string position is valid
            if len(cdst) > 0 then        'destlen > 0: toggle to next selection
               ipos = ipos + isrc          'increment the select-string pointer
            end if
            if ipos > isep then            'pointer is beyond end of selections
               ipos = 1               'reset pointer to beginning of selections
            end if
         else                         'replacement-string position is NOT valid
            ipos = isep + 1                  'set position to select blank data
         end if
         lset cdst = mid$(csep, ipos, isrc)   'put select-data to output buffer
         ilen = ((ipos - 1) \ isrc) * ilen + 1     'replacement-string position
      case "T1"                                       'time type hhmm or hhmmss
         lset cdst = mid$(csrc, 1, 2) + ":" + mid$(csrc, 3, 2) + ":" + _
                     mid$(csrc, 5, 2)               'put data into dest.-buffer
         if left$(cdst, 2) <> "  " and right$(cdst, 2) = "  " then 'source=hhmm
            mid$(cdst, ilen - 1) = "00"                      'add dummy seconds
         end if
      case "T5"                                       'time type ssmh (Btrieve)
         if cvl(csrc) = 0 or rtrim$(csrc) = "" then        'source time is NULL
            lset cdst = "  :  :"                  'set formatted time to blanks
         else                                          'source time is not NULL
            lset cdst = right$("0" + ltrim$(str$(midchar(csrc, 4))), 2) + ":" + _
                        right$("0" + ltrim$(str$(midchar(csrc, 3))), 2) + ":" + _
                        right$("0" + ltrim$(str$(midchar(csrc, 2))), 2)
         end if                             'put data into dest.-buffer [above]
      case else                                          'character-type fields
         lset cdst = left$(csrc, ilen)              'put data into dest.-buffer
   end select
end function

function ibr.hkey(ckey(), ckvl(), cpos(), crcd(), iptr())'get high key in range
   lset ckey(0) = ckey(21)                      'begin at max. allow. key-value
   istat = ibr.btrv(BTRV.GETLE, crcd(1, 0), crcd(0, 0), _
   iptr(4, 0), ckey(0), iptr(7, 0), cpos(24))     'get 1st record <= max. value
   i = ibr.mask(ckey(), ckey(0), ckey(20))     'mask key against min. key-value
   if istat = 0 and ckey(24) >= ckey(25) then      'record found & within range
      lset ckey(23) = ckey(0)                'return highest key-value in range
   else                                   'record NOT found or NOT within range
      lset ckey(23) = ckey(26)                           'return NULL key-value
      lset cpos(24) = cpos(26)                      'return NULL position-value
   end if
end function

function ibr.kbuf(bc as bcnf)           'get key value (don't strip key buffer)
   ltm1 = lbr.time((time$))                     'current time of day in seconds
   do                                               'begin loop to get keypress
      ckey = inkey$                                    'get key from key buffer
      if bc.keyout > 0 and len(ckey) = 0 then   'timeout active and no keypress
         ltm2 = lbr.time((time$)) - ltm1           'current interval in seconds
         if ltm2 >= bc.keyout or ltm2 < 0 and ltm2 >= bc.keyout - 86400 then
            ckey = char(KEY.CSRLT)     'current interval >= key timeout [above]
         end if                           'set key to cursor-left value [above]
      end if
   loop while len(ckey) = 0                            'loop until a key is hit
   if asc(ckey) then                                   'key value in first byte
      ibr.kbuf = asc(ckey)                 'return key value to calling program
   else                                            'zero value ("extended" key)
      ibr.kbuf = asc(mid$(ckey, 2)) + 128           'add 128 to 2nd byte of key
   end if
end function                                         'return to calling program

function ibr.lkey(ckey(), ckvl(), cpos(), crcd(), iptr()) 'get low key in range
   lset ckey(0) = ckey(20)                      'begin at min. allow. key-value
   istat = ibr.btrv(BTRV.GETGE, crcd(1, 0), crcd(0, 0), _
   iptr(4, 0), ckey(0), iptr(7, 0), cpos(23))     'get 1st record >= min. value
   i = ibr.mask(ckey(), ckey(0), ckey(21))     'mask key against max. key-value
   if istat = 0 and ckey(24) <= ckey(25) then      'record found & within range
      lset ckey(22) = ckey(0)                 'return lowest key-value in range
   else                                   'record NOT found or NOT within range
      lset ckey(22) = ckey(26)                           'return NULL key-value
      lset cpos(23) = cpos(26)                      'return NULL position-value
   end if
end function

function ibr.mask(ckey(), cstr1, cstr2)  'reverse binary key segments&uppercase
   i = ibr.uprc(cstr1, ckey(10))               'uppercase 'C'-masked characters
   lset ckey(24) = cstr1                    'put key1 value into compare-buffer
   lset ckey(25) = cstr2                    'put key2 value into compare-buffer
   for i = 2 to 5                             'loop thru the mask-segment types
      select case i                        'select on int, uns., long, uns.long
         case 2                                'short integer type mask segment
            ichr = 73   'I'                 'set type for call to reverse bytes
         case 3                             'unsigned integer type mask segment
            ichr = 85   'U'                 'set type for call to reverse bytes
         case 4                                 'long integer type mask segment
            ichr = 76   'L'                 'set type for call to reverse bytes
         case 5                        'unsigned long integer type mask segment
            ichr = 87   'W'                 'set type for call to reverse bytes
         case else
      end select
      ilen = (i \ 2) * 2                    'set the length of the mask segment
      iptr = 1 - ilen                 'initialize mask-pointer position in mask
      do                                 'begin loop to reverse binary segments
         iptr = instr(iptr + ilen, ckey(10), string$(ilen, ichr))'segm.location
         if iptr then                            'binary segment location found
            for ipos = 0 to ilen - 1           'loop thru the segment positions
               mid$(ckey(24), iptr + ipos) = _ 'put segment byte to key1 buffer
               char(midchar(cstr1, iptr + ilen - ipos - 1))
               mid$(ckey(25), iptr + ipos) = _ 'put segment byte to key2 buffer
               char(midchar(cstr2, iptr + ilen - ipos - 1))
            next
         else                                'binary segment location NOT found
            exit do                      'segment location NOT found; exit loop
         end if
      loop
   next
end function

function ibr.scrl(bc as bcnf, ckvl(), cpos(), cvdb()) 'scroll data-rows up/down
   if bc.scroll < 0 then                                   'scroll data-rows up
      itop = 0                                         'begin scroll at 1st row
      ilow = bc.dsprow - 1                           'end scroll at last row -1
   else                                                  'scroll data-rows down
      itop = bc.dsprow + abs(bc.dsprow < bc.rowcnt - 1) 'begin scroll @last row
      ilow = 1                                           'end scroll at 2nd row
   end if
   for irow = itop to ilow step -sgn(bc.scroll)        'loop thru the data-rows
      if irow > 0 then                                   'loop-row is > 1st row
         icolor = bc.clrdat                            'set color to data-color
      else                                                 'loop-row is 1st row
         icolor = bc.clrbar                      'set color to scroll-bar-color
      end if
      lset ckvl(irow) = ckvl(irow - bc.scroll)       'shift data-row key buffer
      lset cpos(irow) = cpos(irow - bc.scroll)   'shift Btrieve position buffer
      lset cvdb(irow) = cvdb(irow - bc.scroll)     'shift data-row video buffer
      i = io.vdsp(cvdb(irow), bc.rowbeg + irow, bc.colmin + 1, icolor)
   next
end function

function ibr.scrn(bc as bcnf, cdat(), cvdb(), iptr())    'display Browse window
   if bc.linext then                            'OK to clear Browse-window area
      i = io.vcls(bc.rowmin, bc.colmin, bc.rowmax, _
      bc.colmax, bc.linext, 1, bc.clrwin)              'clear the Browse-window
   end if
   if bc.hdrrow then                           'OK to disp. header-title in box
      i = io.vcls(bc.rowmin + 1, bc.colmin + 2, _   'clear the header-title row
      bc.rowmin + 1, bc.colmax - 2, 0, 1, bc.clrhdr)
      i = io.vdsp(cdat(0, 0), bc.rowmin + 1, bc.hdrcol, bc.clrhdr)'disp. header
      i = io.vdsp(bc.cdtl + cvdb(26) + bc.cdtr, _
      bc.rowmin + 2, bc.colmin, bc.clrwin)          'disp. bottom of header-box
   end if
   if bc.titles then                         'titles available; clear title row
      i = io.vcls(bc.rowbeg - bc.titles, bc.colmin + 1, _
      bc.rowbeg - 2, bc.colmax - 1, 0, 1, bc.clrttl)    'clear column-title row
      if bc.linint then                             'OK to disp. internal lines
         i = io.vdsp(bc.cstl + string$(bc.colcnt, 45) _
         + bc.cstr, bc.rowbeg - 1, bc.colmin, bc.clrwin)  'line under title-row
      end if
   end if
   i = io.vcls(bc.rowbeg, bc.colmin + 1, _       'clear Browse-window data area
   bc.rowend, bc.colmax - 1, 0, 1, bc.clrdat)
   if bc.edtmax >= 0 then                        'search field(s) are available
      i = io.vdsp(bc.cstl + string$(bc.colcnt, 45) _
      + bc.cstr, bc.rowend + 1, bc.colmin, bc.clrwin)    'line under search-row
   end if
   if bc.ftrcnt then                           'footer message(s) are available
      i = io.vdsp(bc.cstl + string$(bc.colcnt, 45) + bc.cstr, _
      bc.rowmax - bc.ftrcnt - 1, bc.colmin, bc.clrwin)  'line above footer area
      i = io.vcls(bc.rowmax - bc.ftrcnt, bc.colmin + 2, _
      bc.rowmax - 1, bc.colmax - 2, 0, 1, bc.clrftr)  'clear Browse footer area
   end if
   if bc.titles then                         'titles available; disp. title row
      ifldcol = 1 + bc.colmin + abs(bc.padext)     '1st field (column) position
      for ifld = 0 to bc.fldmax                    'loop thru the Browse fields
         itab = (iptr(2, ifld) - len(cdat(2, ifld))) \ 2 'offset to title-begin
         i = io.vdsp(cdat(2, ifld), bc.rowbeg - _
         bc.titles, ifldcol + itab, bc.clrttl)         'display the field-title
         itab = (iptr(2, ifld) - len(cdat(3, ifld))) \ 2 'offs. 2nd-title-begin
         i = io.vdsp(cdat(3, ifld), bc.rowbeg - 2, ifldcol + itab, bc.clrttl)
         i = io.vdsp(string$(iptr(2, ifld), 45), _
         bc.rowbeg - 1, ifldcol, bc.clrwin)        'line under each title-field
         ifldcol = ifldcol + iptr(2, ifld)      'move to field-separator column
         if ifld < bc.fldmax and bc.linint = 1 then 'no separ.after final field
            i = io.vdsp(bc.cstu, bc.rowbeg - bc.titles - 1, ifldcol, bc.clrwin)
            i = io.vdsp("|", bc.rowbeg - bc.titles, ifldcol, bc.clrwin)
            i = io.vdsp("|", bc.rowbeg - 2, ifldcol, bc.clrwin)'vert. separator
            i = io.vdsp("-", bc.rowbeg - 1, ifldcol, bc.clrwin)'line-join char.
            for irow = bc.rowbeg to bc.rowend  'loop thru the data-display rows
               i = io.vdsp("|", irow, ifldcol, bc.clrdat)'separator in data row
            next
            for irow = 0 to 24                 'loop thru the data-display rows
               mid$(cvdb(irow), ifldcol - bc.colmin) = "|" 'put field-separator
            next
            if bc.edtmax >= 0 then               'search field(s) are available
               i = io.vdsp("-", bc.rowend + 1, ifldcol, bc.clrwin) 'line-joiner
               i = io.vdsp("|", bc.rowend + 2, ifldcol, bc.clrwin)'edit-row sep
            end if
            if bc.ftrcnt then                  'footer message(s) are available
               i = io.vdsp("-", bc.rowend + abs(bc.edtmax _
               >= 0) * 2 + 1, ifldcol, bc.clrwin)   'line-joiner under edit-row
            else                           'footer message(s) are NOT available
               i = io.vdsp(bc.cstd, bc.rowmax, ifldcol, bc.clrwin) 'line-joiner
               mid$(cvdb(26), ifldcol - bc.colmin) = bc.cstd   'field-separator
            end if
         end if
         ifldcol = ifldcol + 1               'move to next field-display column
      next
   end if
   if not bc.hdrrow then         'OK to disp. header-title on top window-border
      i = io.vdsp(cdat(0, 0), bc.rowmin, bc.hdrcol, bc.clrhdr)  'display header
   end if
   irow = bc.rowmax - bc.ftrcnt - 1        'initialize row# for footer-messages
   for indx = 3 to 5                             'loop thru the footer-messages
      if len(cdat(0, indx)) then            'footer-message exists @ this index
         irow = irow + 1                    'increment row# when message exists
         i = io.vdsp(cdat(0, indx), irow, bc.ftrcol, bc.clrftr)'display message
      end if
   next
end function

function lbr.time(ctim)              'return the current time of day in seconds
   ihrs = midchar(ctim, 1) * 10 + midchar(ctim, 2)          'current-time hours
   imin = midchar(ctim, 4) * 10 + midchar(ctim, 5)        'current-time minutes
   isec = midchar(ctim, 7) * 10 + midchar(ctim, 8)        'current-time seconds
   lbr.time = clng(ihrs) * 3600 + imin * 60 + isec  'total current-time seconds
end function

function ibr.uprc(cstr1, cmask)                'uppercase 'C'-masked characters
   for imskptr = 1 to len(cmask)                  'loop thru the masking string
      if midchar(cmask, imskptr) = 67 then               'position to uppercase
         ichr = midchar(cstr1, imskptr)          'ASCII value at char. position
         if ichr >= 97 and ichr <= 122 then      'char. at imskptr is lowercase
            mid$(cstr1, imskptr) = char(ichr - 32)  'convert char. to uppercase
         end if
      end if
   next
end function
