'/* MCRE.BAS Make a '.CRE' file from a BUTIL -stat file */
'/*          By: Dale Thorn                             */
'/*          Rev. 20.04.2000                            */

'$include: 'basdef.h'
'$include: 'filekill.h'
'$include: 'fileopen.h'
'$include: 'longname.h'
'$include: 'messages.h'
'$include: 'midchar.h'
'$include: 'parmstr1.h'
'$include: 'string.h'
'$include: 'basdef.bas'
'$include: 'filekill.bas'
'$include: 'fileopen.bas'
'$include: 'longname.bas'
'$include: 'messages.bas'
'$include: 'midchar.bas'
'$include: 'parmstr1.bas'
'$include: 'string.bas'
'$include: 'scrnparm.bas'

ccmd = ucase$(rtrim$(command$))         'get the user's command-line parameters
if ccmd = "" then                              'a command line was NOT supplied
   i = ifn.msgs("Usage:  MCRE  filename", iofs, irow, icol, 0, 1)
end if                              'display the usage message [above] and exit

iprm = parmstr1(ccmd, cfil, cnam, cext, cprm())  'parse command-line parameters
if cnam = "" or len(cnam) > 8 or len(cext) > 3 or instr(cext, ".") _
or iprm >= 0 then                    'invalid filename or additional parameters
   i = ifn.msgs("Invalid filename", iofs, irow, icol, 1, 1)
end if                       'display invalid-filename message [above] and exit

i = ifn.open(1, cfil, "B", llof)           'open the source file in binary mode
if llof < 0 then                                'user input a wildcard filespec
   i = ifn.msgs("Invalid filename", iofs, irow, icol, 1, 1)      'beep and exit
elseif llof = 0 then                          'source file nonexistent or empty
   i = ifn.kill(1, cfil)                             'kill the zero-length file
   i = ifn.msgs(cfil + " not found", iofs, irow, icol, 1, 1)     'beep and exit
end if

i = ifn.msgs("Please standby", iofs, irow, icol, 0, 0)         'standby message

close 1                                   'close the binary-access file channel
inosegs = 0                               'initialize the total no. of segments
open cfil for input as 1                                  'open the source file
do while not eof(1)                            'loop thru the source-file lines
   line input #1, clin                        'read a line from the source file
   clin = ucase$(ltrim$(rtrim$(clin)))      'trim and uppercase the source line
   if left$(clin, 19) = "FILE STATISTICS FOR" then       'Btrieve filename line
      for ipos = len(clin) to 21 step -1         'walk to beginning of filename
         if midchar(clin, ipos) = 92 then      '\' character found in text line
            exit for                   'beginning of filename found - exit loop
         end if
      next
      cnam = mid$(clin, ipos + 1)         'set base name for '.CRE' script file
      idot = instr(cnam, ".")              'filename extension position in cnam
      if idot = 0 then                    'filename extension NOT found in cnam
         idot = len(cnam) + 1            'set extension position to end of cnam
      end if
      cdst = left$(cnam, idot - 1) + ".CRE"    'set name for '.CRE' script file
   elseif left$(clin, 11) = "PAGE SIZE =" then  'pagesize from BUTIL -stat file
      ipagsiz = pdqvali(mid$(clin, 12))            'get pagesize from text line
      if ipagsiz > 2048 then                     'pagesize > max.for Btrieve v5
         ipagsiz = 2048                         'set pagesize to Btrieve v5 max.
      end if
   elseif left$(clin, 15) = "RECORD LENGTH =" then   'recordlen from BUTIL file
      ibuflen = pdqvali(mid$(clin, 16))           'get recordlen from text line
   elseif left$(clin, 18) = "VARIABLE RECORDS =" then 'var.rcds.from BUTIL file
      cvar = mid$(clin, 20, 1)                    'get var.rcds. from text line
   elseif left$(clin, 26) =  "TOTAL NUMBER OF SEGMENTS =" then  '#segments line
      inosegs = pdqvali(mid$(clin, 27))     'get no. of segments from text line
   end if
   if inosegs then                          'no. of segments value has been set
      if left$(clin, 7) = "SEGMENT" then   'the next line is a key segment line
         exit do              'exit this loop and process the key segment lines
      end if
   end if
loop

open cdst for output as 2                          'open the '.CRE' script file
print #2, "'/* "; cnam; " creation script file */"  'hdr.for '.CRE' script file
print #2, ""                          'separator line for next ('FILE') section
print #2, "* ID  LEN  VAR  PAGE"                     'header for 'FILE' section
print #2, "FILE "; right$("   " + str$(ibuflen), 4); "    "; _  'buffer length,
             cvar; right$("   " + str$(ipagsiz), 6)  'variable length, pagesize
print #2, ""                           'separator line for next ('KEY') section
print #2, "* ID  KEY  SEG  POS  LEN  DUP  MOD  ALT  TYPE  NULL"'hdr.for KEY sec

for ilin = 1 to inosegs                      'loop thru the no. of key segments
   line input #1, clin                        'read a line from the source file
   clin = ucase$(ltrim$(rtrim$(clin)))      'trim and uppercase the source line
   iptr = 1                                                'key number position
   ipos = istr.lcfn(iptr, clin, " ")                'space following key number
   ckey = mid$(clin, iptr, ipos - iptr)               'key number (string form)
   iptr = istr.lcsp(ipos, clin, " ")                  'key segment no. position
   ipos = istr.lcfn(iptr, clin, " ")               'space following key segment
   cseg = ltrim$(str$(pdqvali(mid$(clin, iptr, ipos - iptr)) - 1)) 'key segment
   iptr = istr.lcsp(ipos, clin, " ")                'key segment begin position
   ipos = istr.lcfn(iptr, clin, " ")             'space following segment begin
   cbeg = mid$(clin, iptr, ipos - iptr)        'key segment begin (string form)
   iptr = istr.lcsp(ipos, clin, " ")               'key segment length position
   ipos = istr.lcfn(iptr, clin, " ")            'space following segment length
   clen = mid$(clin, iptr, ipos - iptr)       'key segment length (string form)
   iptr = istr.lcsp(ipos, clin, " ")                 'key segment type position
   ipos = istr.lcfn(iptr, clin, " ")              'space following segment type
   SELECT CASE mid$(clin, iptr, 8)             'key segment type (7 chars.max.)
      CASE "AUTOINC "                           'autoincrement type key segment
         ctyp = "A"                             'set autoincrement type segment
      CASE "BFLOAT  "                         'BASIC MBF float type key segment
         ctyp = "B"                           'set BASIC MBF float type segment
      CASE "DATE    "                            'Btrieve date type key segment
         ctyp = "D"                              'set Btrieve date type segment
      CASE "FLOAT   "                        'BASIC IEEE float type key segment
         ctyp = "F"                          'set BASIC IEEE float type segment
      CASE "INTEGER ", _                        'BASIC Integer type key segment
           "LOGICAL "                           'BASIC Integer type key segment
         if clen = "2" then            'length==2 indicates BASIC short Integer
            ctyp = "I"                    'set BASIC short Integer type segment
         else                           'length==4 indicates BASIC long Integer
            ctyp = "L"                     'set BASIC long Integer type segment
         end if
      CASE "STRING  "                            'BASIC String type key segment
         ctyp = "S"                              'set BASIC String type segment
      CASE "TIME    "                            'Btrieve time type key segment
         ctyp = "T"                              'set Btrieve time type segment
      CASE "UNSIGNED"                            'BASIC binary type key segment
         if clen = "2" then            'length==2 indicates BASIC short Integer
            ctyp = "U"                    'set BASIC short Integer type segment
         else                           'length==4 indicates BASIC long Integer
            ctyp = "W"                     'set BASIC long Integer type segment
         end if
      CASE ELSE                                   'unspecified type key segment
         i = ifn.msgs("Invalid field type: " + mid$(clin, iptr, ipos - iptr), _
                      iofs, irow, icol, 1, 1)   'display error message and exit
   END SELECT
   iptr = iptr + 8                                    'key flags begin position
   ipos = iptr + 8                             'position after end of key flags
   if instr(mid$(clin, iptr, ipos - iptr), "<") then   'key is descending order
      cdes = "-"                                        'set descending type ON
   else                                                'key NOT descending type
      cdes = " "                                       'set descending type OFF
   end if
   if instr(mid$(clin, iptr, ipos - iptr), "I") then   'key is case-insensitive
      calt = "Y"                                       'set case-insensitive ON
   else                                               'key NOT case-insensitive
      calt = "N"                                      'set case-insensitive OFF
   end if
   if instr(mid$(clin, iptr, ipos - iptr), "M") then  'modifiable type position
      cmod = "Y"                                        'key is modifiable type
   else                                                 'set modifiable type ON
      cmod = "N"                                         'key is NOT modifiable
   end if                                              'set modifiable type OFF
   if instr(mid$(clin, iptr, ipos - iptr), "D") then       'key is duplicatable
      cdup = "Y"                                      'set duplicatable type ON
   else                                                'key is NOT duplicatable
      cdup = "N"                                     'set duplicatable type OFF
   end if
   iptr = istr.lcsp(ipos, clin, " ")                    'null-key type position
   ipos = istr.lcfn(iptr, clin, " ")                 'space following null type
   if pdqvali(mid$(clin, iptr, ipos - iptr)) = 20 then      'hex 20 == ASCII 32
      cnul = "32"                              'set ASCII 32 (spaces) null type
   else                                                 'key is NOT a null type
      cnul = "--"                                        'set null-key type OFF
   end if
   iptr = istr.lcsp(ipos, clin, " ")                   'no. of records position
   ipos = istr.lcfn(iptr, clin, " ")                  'space following #records
   iptr = istr.lcsp(ipos, clin, " ")               'collating-sequence position
   ipos = istr.lcfn(iptr, clin, " ")              'space following collate-spec
   if calt = "N" then                         'case sensitivity NOT already set
      if ltrim$(mid$(clin, iptr, ipos - iptr)) = "0" then  '0': uprc. collating
         calt = "Y"                                 'set uppercase collating ON
      else                                          'key NOT uppercase collated
         calt = "N"                                'set uppercase collating OFF
      end if
   end if
   print #2, "KEY    "; _                                   'key segment header
             right$(" "  + ckey, 2); "   "; _                       'key number
             right$(" "  + cseg, 2); "  "; _                    'segment number
             right$("00" + cbeg, 3); "  "; _          'key segment buffer begin
             right$("00" + clen, 3); "    "; _       'key segment buffer length
             cdup; "    "; _                     'key segment duplicatable type
             cmod; "    "; _                       'key segment modifiable type
             calt; "    "; _                          'key segment alt-key type
             cdes; _                               'key segment descending type
             ctyp; "    "; _                             'key segment data type
             cnul                                        'key segment null type
next

i = ifn.msgs(cdst + " created successfully", iofs, irow, icol, 0, 1)'msg.& exit
close                                 'close all files in case not closed above
system                                      'return control to operating system
