'/* TEXTSORT.BAS  Sort a text file with/without pointers */
'/*               By: Dale Thorn                         */
'/*               Rev. 04.02.1998                        */

'$include: 'basdef.h'
'$include: 'filekill.h'
'$include: 'fileopen.h'
'$include: 'messages.h'
'$include: 'midchar.h'
'$include: 'parmstr1.h'
'$include: 'tsort.h'

declare function io.kbuf(inop)
declare function io.kget(inop)

'$include: 'basdef.bas'
'$include: 'filekill.bas'
'$include: 'fileopen.bas'
'$include: 'messages.bas'
'$include: 'midchar.bas'
'$include: 'parmstr1.bas'
'$include: 'scrnparm.bas'
'$include: 'tsort.bas'

ccmd = ucase$(rtrim$(command$))            'get user's command-line parameters
if ccmd = "" then                   'command-line parameters were NOT supplied
   i = ifn.msgs("Usage:  TEXTSORT  filename  [offset]  [max.wordlen]", _
   iofs, irow, icol, 0, 1)             'display usage message [above] and exit
end if

iprm = parmstr1(ccmd, cfil, cnam, cext, cprm()) 'parse command-line parameters
if cnam = "" or len(cnam) > 8 or len(cext) > 3 or instr(cext, ".") then
   i = ifn.msgs("Invalid filename", iofs, irow, icol, 1, 1)  'invalid filename
end if                      'display invalid-filename message [above] and exit
if iprm > 1 then                         'no. of parameters was > 1 (too many)
   i = ifn.msgs("Invalid number of parameters", iofs, irow, icol, 1, 1)
end if                     'display no.-of-parameters message [above] and exit
if iprm = 1 then                        'a max. wordlen parameter was supplied
   imaxlen = abs(pdqvali(cprm(1)))    'numeric value of max. wordlen parameter
   if imaxlen < 2 or imaxlen > 35 then         'invalid max. wordlen parameter
      i = ifn.msgs("Invalid max. wordlen parameter", iofs, irow, icol, 1, 1)
   end if                  'display invalid-parameter message [above] and exit
else                                'a max. wordlen parameter was NOT supplied
   imaxlen = 35                            'set the default max. wordlen to 35
end if
cprm(1) = ltrim$(str$(imaxlen))       'max. wordlen parameter (character form)
if iprm >= 0 then                            'an offset parameter was supplied
   ioffset = abs(pdqvali(cprm(0)))          'numeric value of offset parameter
   if ioffset > 35 then                          'invalid offset specification
      i = ifn.msgs("Invalid offset parameter", iofs, irow, icol, 1, 1)
   end if                  'display invalid-parameter message [above] and exit
else                                     'an offset parameter was NOT supplied
   ioffset = 0                                 'set the default offset to zero
end if
cprm(0) = ltrim$(str$(ioffset))             'offset parameter (character form)

nc.sdno = 1                      'file channel/unit# for sort data file (.sdx)
nc.ndno = 2                     'file channel/unit# for sort index file (.ndx)
nc.ddno = 3                   'file channel/unit# for output index file (.ddx)
iauxfil = 4                                'auxiliary file-unit/channel number

csrc = cnam + ".pts"                        'add extension to word-source file
cdst = cnam + ".ptr"                      'add extension to pointer/index file
i = ifn.kill(iauxfil, "wmak.err")                    'kill the word-error file
i = ifn.msgs("Creating wordfile - Please standby", iofs, irow, icol, 0, 0)
shell "wmak " + cfil + " " + cprm(0) + " " + cprm(1)'make the word-source file
open "wmak.err" for binary as iauxfil                'open the word-error file
caux = space$(1)                                   'buffer for word-error file
if lof(iauxfil) then                    'an error code was written by WMAK.EXE
   get iauxfil, 1, caux                             'get the error-status byte
end if
i = ifn.kill(iauxfil, "wmak.err")                    'kill the word-error file
select case caux                                    'error codes from WMAK.EXE
   case " ", _                       'WMAK.EXE failed for undetermined reasons
        "9"
      i = ifn.msgs("WMAK program failed or was not available", iofs, irow, icol, 1, 1)
   case "0"                          'normal processing successfully completed
      print                             'print a blank line before the message
      print tab(iofs); "Hit any key to continue; Esc to quit"   'user-continue
   case "1"                                       'ASCII source file NOT found
      i = ifn.msgs(cfil + " not valid or not found", iofs, irow, icol, 0, 1)
   case "2"                                 'source file found; no valid words
      i = ifn.kill(iauxfil, csrc)                   'kill the word-source file
      i = ifn.msgs("No valid words were found in " + cfil, iofs, irow, icol, 0, 1)
   case "3"                                     'WMAK failed memory allocation
      i = ifn.msgs("Insufficient memory to run WMAK program", iofs, irow, icol, 0, 1)
   case else
end select
ikey = io.kget(0)                                    'get the user-pressed key
if ikey = 27 then                                         'Esc key was pressed
   i = ifn.msgs("Abort selected - Program terminated", iofs, irow, icol, 0, 1)
else
   i = ifn.msgs("Sorting wordfile - Please standby", iofs, irow, icol, 0, 0)
end if

i = ifn.open(iauxfil, csrc, "B", llof)     'open the source file in binary mode
if llof < 0 then                                'user input a wildcard filespec
   i = ifn.msgs("Invalid filename", iofs, irow, icol, 1, 1)      'beep and exit
elseif llof = 0 then                          'source file nonexistent or empty
   i = ifn.kill(iauxfil, csrc)                       'kill the zero-length file
   i = ifn.msgs(csrc + " not found", iofs, irow, icol, 1, 1)     'beep and exit
end if
cnum = space$(4)                             'buffer for total number of words
get iauxfil, llof - 3, cnum                        'fill the word-total buffer
cnum = ltrim$(str$(cvl(cnum)))             'character form of word-total value
caux = space$(2)                               'buffer for maximum word length
get iauxfil, llof - 5, caux                       'fill the word-length buffer
iwrdlen = cvi(caux)                               'get the maximum word length
nc.slen = iwrdlen + 3                                'total sort buffer length
caux = space$(nc.slen + 1)                          'source-file record buffer
cnc.sdat = space$(nc.slen)                            'sort data record buffer
insbeg(0) = 1                       'segment begin pointer for cnc.sdat buffer
inslen(0) = nc.slen                                            'segment length
insseq(0) = 1                    'segment sort sequence (ascending/descending)
nc.oper = 0               'set flag to add records to sort (initial operation)
lptr = 1                         'initialize the source-file data-read pointer

locate irow - 1, 1, 0                   'position cursor for following message
print " Sorting "; space$(len(cnum)); " of "; cnum; " words";    'word-counter
do while lptr < llof - 5                             'loop to add data to sort
   get iauxfil, lptr, caux                        'load the source-file buffer
   ilen = midchar(caux, 1)                         'length of the current word
   lset cnc.sdat = mid$(caux, 2, ilen)                   'load the sort-buffer
   mid$(cnc.sdat, iwrdlen + 1) = char(midchar(caux, ilen + 4))       'swap the
   mid$(cnc.sdat, iwrdlen + 2) = char(midchar(caux, ilen + 3))        'pointer
   mid$(cnc.sdat, iwrdlen + 3) = char(midchar(caux, ilen + 2))          'bytes
   i = psort(nc, cnc.data, cnc.indx, cnc.lng, cnc.nbuf, _'add data to the sort
   cnc.sdat, insbeg(), inslen(), insseq(), lnsptr())
   lptr = lptr + ilen + 4         'increment the source-file data-read pointer
   if (nc.rptr - 1) mod 100 = 0 then     'another 100 records have been sorted
      locate irow - 1, 10, 0            'position cursor for following message
      print ltrim$(str$(nc.rptr - 1));      'display the current counter value
   end if
loop

nc.drow = irow          'pass the screen bottom-row number to the sort routine
nc.oper = -2         'set flag to retrieve records from sort (final operation)
i = psort(nc, cnc.data, cnc.indx, cnc.lng, cnc.nbuf, _
cnc.sdat, insbeg(), inslen(), insseq(), lnsptr())'call sort to build the index

i = ifn.kill(nc.ddno, cdst)                                'kill the .PTR file
name "sortdata.ddx" as cdst          'set the sort-index to the .PTR file name
open cdst for binary as nc.ddno                            'open the .PTR file
llof = lof(nc.ddno)                        'set to the length of the .PTR file
cnc.nbuf = mki$(ioffset)                         'make the offset-value buffer
put nc.ddno, llof + 1, cnc.nbuf                'add offset to end of .PTR file
lset cnc.nbuf = mki$(iwrdlen)                   'make the wordlen-value buffer
put nc.ddno, llof + 3, cnc.nbuf           'add word-length to end of .PTR file

i = ifn.msgs("Word sort completed", iofs, irow, icol, 1, 1)      'User-message
close                                                         'close all files
system                                                          'return to DOS

function io.kbuf(inop)            '/* get key value (don't strip key buffer) */
   do                                                         '/* begin loop */
      ckey = inkey$                              '/* get key from key buffer */
   loop while ckey = ""                          '/* loop until a key is hit */
   if asc(ckey) then                             '/* key value in first byte */
      io.kbuf = asc(ckey)            '/* return key value to calling program */
   else                                      '/* zero value ("extended" key) */
      io.kbuf = asc(mid$(ckey, 2)) + 128      '/* add 128 to 2nd byte of key */
   end if
end function                                   '/* return to calling program */

function io.kget(inop)                 '/* strip key buffer and wait for key */
   do                                                         '/* begin loop */
   loop while len(inkey$)           '/* loop while key buffer contains a key */
   io.kget = io.kbuf(0)              '/* return key value to calling program */
end function                                   '/* return to calling program */
