'/* TRIMPRT.BAS Reformat columns in .PRT file to set optimum width */
'/*             By: Dale Thorn                                     */
'/*             Rev. 05.12.2002                                    */

'$include: 'basdef.h'
'$include: 'filekill.h'
'$include: 'fileopen.h'
'$include: 'longname.h'
'$include: 'messages.h'
'$include: 'midchar.h'
'$include: 'parmstr1.h'
'$include: 'string.h'
'$include: 'basdef.bas'
'$include: 'filekill.bas'
'$include: 'fileopen.bas'
'$include: 'longname.bas'
'$include: 'messages.bas'
'$include: 'midchar.bas'
'$include: 'parmstr1.bas'
'$include: 'string.bas'

dim ifldlen(22)                             'initialize the field-lengths array
dim ifldsiz(22)                           'initialize the initial lengths array
dim inumfmt(22)                            'initialize the numeric fields flags

ccmd = ucase$(rtrim$(command$))             'get the user's command-line params
if ccmd = "" then                              'command-line params NOT entered
   cls                                            'clear the DOS display screen
   locate 5, 1, 1                                   'locate cursor for messages
   print "    Usage:  TRIMPRT  filename"             'display the usage message
   print
   print "            TRIMPRT  filename  [/M##]"    'display the option message
   print
   print
   print "    If '/M' specified, set minimum column width to '##'"
   print "                       [## = 2 to 99]"    'display add'l info message
   system                                             'exit to operating system
end if

if left$(right$(ccmd, 3), 2) = "/M" then      'OK to set a minimum column width
   iminlen = pdqvali(right$(ccmd, 1))         'set min. width to user selection
   ccmd = rtrim$(left$(ccmd, len(ccmd) - 3))   'remove '/M#' user specification
elseif left$(right$(ccmd, 4), 2) = "/M" then  'OK to set a minimum column width
   iminlen = pdqvali(right$(ccmd, 2))         'set min. width to user selection
   ccmd = rtrim$(left$(ccmd, len(ccmd) - 4))   'remove '/M#' user specification
else                                          'minimum column width setting n/a
   iminlen = 0                                'set minimum column width to zero
end if

iprm = parmstr1(ccmd, cfil, cnam, cext, cprm())  'parse command-line parameters
if iprm >= 0 then                                 'invalid number of parameters
   i = ifn.msgs("Invalid number of parameters", 5, 24, 79, 1, 1)
end if                                  'display error message [above] and exit

i = ifn.open(1, cfil, "S", llof)           'open source file in sequential mode
if llof < 0 then                                'user input a wildcard filespec
   i = ifn.msgs("Invalid filename", 5, 24, 79, 1, 1) 'display error msg. & exit
elseif llof = 0 then                          'source file nonexistent or empty
   i = ifn.kill(1, cfil)                             'kill the zero-length file
   i = ifn.msgs(cfil + " not found", 5, 24, 79, 1, 1)'display error msg. & exit
end if                                  'display error message [above] and exit

if cext = "out" then                        'the source file extension is '.OUT'
   cout = cnam + ".tmp"                      'set dest.-file extension to '.TMP'
else                                        'source file extension is NOT '.OUT'
   cout = cnam + ".out"                      'set dest.-file extension to '.OUT'
end if

i = ifn.msgs("Please standby", 5, 24, 79, 0, 0)              'OK to proceed....

init = 0                         'initialize flag to identify first detail line
do while not eof(1)                 'loop until end of the source ('.PRT') file
   line input #1, csrclin             'get next available line from source file
   if rtrim$(csrclin) <> "" then          'the current source line is non-blank
      if not init then                     'currently processing header section
         if left$(csrclin, 2) = "--" then 'last header line before detail lines
            init = not 0                'set to ID next line as 1st detail line
            isrcpos = 1                 'initialize begin position in text line
            ifldmax = -1                 'initialize max. field no. array index
            ilinlen = len(rtrim$(csrclin))  'end of non-blank data in text line
            do while isrcpos <= ilinlen       'process fields until end of line
               idstpos = instr(isrcpos, csrclin, " ")     'next field separator
               if idstpos = 0 then              'next field separator NOT found
                  idstpos = ilinlen + 1      'set field end posn.to end of line
               end if
               ifldmax = ifldmax + 1         'increment to current field number
               ifldsiz(ifldmax) = idstpos - isrcpos    'set original field size
               isrcpos = idstpos + 1        'increment to next field begin posn.
            loop
         end if
      else                                 'currently processing detail section
         isrcpos = 1                    'initialize begin position in text line
         for ifldnum = 0 to ifldmax       'loop through fields identified above
            itxtlen = len(ltrim$(rtrim$(mid$(csrclin, isrcpos, ifldsiz(ifldnum)))))
            if itxtlen then                    'current field data is non-blank
               if inumfmt(ifldnum) = 0 then 'current field NOT yet ID'd numeric
                  if mid$(csrclin, isrcpos, 1) = " " then '1st byte is blank...
                     inumfmt(ifldnum) = -1          '...so this must be numeric
                  end if
               end if
               if inumfmt(ifldnum) <> -2 then'current field NOT yet ID'd string
                  if instr(" -.0123456789", mid$(csrclin, isrcpos, 1)) = 0 then
                     inumfmt(ifldnum) = -2           '...so this must be string
                  end if
               end if
               if ifldlen(ifldnum) < itxtlen then  'current data is longest yet
                  ifldlen(ifldnum) = itxtlen       'reset to longest data found
               end if
               if ifldlen(ifldnum) < iminlen then  'current length l.t. minimum
                  ifldlen(ifldnum) = iminlen     'reset field to minimum length
               end if
            end if
            isrcpos = isrcpos + ifldsiz(ifldnum) + 1  'incr. to next field posn.
         next
      end if
   end if
loop                            'end of initial processing to set field lengths
close                          'close the source ('.PRT') file and reopen below

ilinlen = 0                       'initialize length of destination output line
for ifldnum = 0 to ifldmax         'loop through fields in source ('.PRT') file
   if ifldlen(ifldnum) then      'field has non-blank data in one or more lines
      ilinlen = ilinlen + ifldlen(ifldnum) + 2 'sum field and separator lengths
   end if
next
cdstlin = space$(ilinlen - 2)   'initialize output line (don't pad final field)

open cfil for input as 1           'reopen source file in sequential input mode
open cout for output as 2            'open dest. file in sequential output mode
do while not eof(1)                     'loop until end of source ('.PRT') file
   line input #1, csrclin                'get next avail. line from source file
   if rtrim$(csrclin) <> "" then          'the current source line is non-blank
      isrcpos = 1                       'initialize begin position in text line
      idstpos = 1                     'initialize begin position in output line
      lset cdstlin = ""                 'clear output line of any existing data
      for ifldnum = 0 to ifldmax        'loop through fields in the source file
         if ifldlen(ifldnum) then       'field has non-blank data in some lines
            cfldtxt = ltrim$(rtrim$(mid$(csrclin, isrcpos, ifldsiz(ifldnum))))
            itxtlen = len(cfldtxt)         'actual text length in current field
            if itxtlen > ifldlen(ifldnum) then  'text is longer than field size
               mid$(cdstlin, idstpos) = left$(cfldtxt, ifldlen(ifldnum)) 'trunc.
            else                           'text length is l.e. to field length
               if inumfmt(ifldnum) = -1 then  'current field is a numeric field
                  ioffset = ifldlen(ifldnum) - itxtlen 'for right-justification
               else                          'current field NOT a numeric field
                  ioffset = 0              'current data will be left-justified
               end if
               if idstpos + ioffset <= len(cdstlin) then   'write inside line!!
                  mid$(cdstlin, idstpos + ioffset) = cfldtxt'add data to output
               end if
            end if
            idstpos = idstpos + ifldlen(ifldnum) + 2  'incr.posn.in output line
         end if
         isrcpos = isrcpos + ifldsiz(ifldnum) + 1     'incr.posn.in source line
      next
      print #2, rtrim$(cdstlin)               'write output line to output file
   end if
loop
close                          'close the source and destination (output) files

shell "brow " + cout              'browse (display) the destination output file
close                                 'close all files in case not closed above
system                                  'return control to the operating system
