'/* TSORT.BAS  Sort/retrieve/index data; ascending/descending; mixed types */
'/*            By: Dale Thorn                                              */
'/*            Rev. 04.02.1998                                             */

dim insbeg(10), inslen(10), insseq(10), lnsptr(1300, 1)
dim nc as ncnf                                     'sort-options pass structure


function psort(nc as ncnf, cnc.data, cnc.indx, cnc.lng, _ 'sort & retrieve data
cnc.nbuf, cnc.sdat, insbeg(), inslen(), insseq(), lnsptr())
   if nc.oper = 0 then                       'insert a record <add to the sort>
      if nc.init = 0 then                     'first-sort-record initialization
         nc.init = -1                               'adjust initialization flag
         nc.sseq = insseq(0)                           'primary output sequence
         i = ifn.kill(nc.sdno, "sortdata.sdx")    'kill work data file (if any)
         i = ifn.kill(nc.ndno, "sortdata.ndx")   'kill work index file (if any)
         i = ifn.kill(nc.ddno, "sortdata.ddx") 'kill output index file (if any)
         open "sortdata.sdx" for binary as nc.sdno         'open work data file
         open "sortdata.ndx" for binary as nc.ndno        'open work index file
         open "sortdata.ddx" for binary as nc.ddno    'open output pointer file
         cnc.lng = string$(4, 0)           'buffer stores a long-integer string
         lfree = lps.free(4096, clng(195840))              'reserve 4 kb memory
         nc.rmax = lfree \ (nc.slen + 6)      'maximum records per memory group
         if nc.rmax > 32640 \ nc.slen then   'records available > string-length
            nc.rmax = 32640 \ nc.slen          'reset maximum records per group
         end if
         cnc.data = space$(nc.rmax * nc.slen)            'main sort data buffer
         cnc.indx = space$(nc.rmax * 3 + 3) 'reorderable/shiftable index buffer
         nc.rptr = 1                               'used to count total records
         nc.rcnt = 1                 'used to count records within a sort group
         nc.gptr = 1                                       'sort buffer pointer
      end if
      if nc.rcnt > nc.rmax then              'too many records to fit in memory
         if nc.ntot = 0 then     'first group; initialize index group variables
            nc.gtot = nc.rcnt - 1                  'number of records per group
            nc.nlen = nc.gtot * 3                    'size of index file buffer
         end if                             'put group data to the sort [below]
         i = ips.ssav(nc, cnc.data, cnc.indx, insbeg(), inslen(), insseq())
         nc.offs = nc.rptr - 1                  'group-to-record offset counter
         nc.rcnt = 1                                'reset group record counter
         nc.gptr = 1                                       'sort buffer pointer
      end if
      i = ips.fill(nc, cnc.data, cnc.indx, cnc.lng, _   'add record to the sort
      cnc.sdat, insbeg(), inslen(), insseq())
      nc.rptr = nc.rptr + 1                    'increment total records counter
      nc.rcnt = nc.rcnt + 1                     'increment group record counter
      nc.gptr = nc.rcnt                                    'sort buffer pointer
   else                                    'retrieve a record or build an index
      nc.offs = 0                               'group-to-record offset counter
      if nc.init = -1 then               'first retrieval record initialization
         nc.init = -2                               'adjust initialization flag
         nc.rtot = nc.rptr - 1                'total records from original sort
         nc.stot = nc.rcnt - 1       'partial-group subtotal from original sort
         nc.dptr = 1                         'beginning pointer for data output
         if nc.ntot then       'sorting was done in groups; put to sort [below]
            i = ips.ssav(nc, cnc.data, cnc.indx, insbeg(), inslen(), insseq())
            igrplen = nc.nlen                                   'group size * 3
            cnc.data = ""                       'erase buffer to reclaim memory
            cnc.indx = ""                       'erase buffer to reclaim memory
            cnc.data = space$(nc.ntot * nc.slen)   'buffer holds 1 record/group
            cnc.indx = space$(nc.ntot * 3 + 3)     'buffer holds 1 record/group
         else                                   'all sorting was done in memory
            nc.dtot = nc.rtot                          'total records to output
            nc.rmax = nc.rtot             'reset maximum records for file write
            nc.nlen = nc.rmax * 3                'length of index data to write
            i = ips.ssav(nc, cnc.data, cnc.indx, insbeg(), inslen(), insseq())
            cnc.data = ""                       'erase buffer to reclaim memory
            cnc.indx = ""                       'erase buffer to reclaim memory
            nc.ntot = 0                   'reset to zero after call to ips.ssav
         end if
         iptrlen = 300                     'length of the pointer-output buffer
         cnc.ptr = space$(iptrlen)        'initialize the pointer-output buffer
         lfree = lps.free(4096, clng(32640))               'reserve normal 4 kb
         lsize = nc.rtot * 3                                 'total records * 3
         nc.mndx = (lsize <= 32640 and lsize <= lfree)    'index-in-memory flag
         if nc.mndx then               'retrieval index fits entirely in memory
            nc.nlen = lsize                 'buffer length is index file length
            cnc.nbuf = space$(nc.nlen)              'allocate index file buffer
            i = ips.rget(nc.ndno, nc.nlen, clng(1), cnc.nbuf)   'load the index
         else                           'retrieval index does not fit in memory
            nc.nlen = 3                 'buffer length is 24-bit integer length
            cnc.nbuf = space$(nc.nlen)              'allocate index file buffer
         end if
         if nc.ntot then                             'merge-sort initialization
            ixx1 = (nc.sseq > 0)                  'used locally to shorten line
            ixx2 = (nc.sseq < 0)                  'used locally to shorten line
            ixx3 = (nc.mndx * 2 and ixx1)         'used locally to shorten line
            ixx4 = (nc.mndx * 2 and ixx2)         'used locally to shorten line
            iyy1 = 1 - nc.mndx * 2                'used locally to shorten line
            iyy2 = igrplen \ (1 - (not nc.mndx) * 2) 'used here to shorten line
            for ircdcnt = 1 to nc.ntot              'loop thru each index group
               nc.rcnt = ircdcnt                 'set sort group record pointer
               nc.gptr = ircdcnt                 'set sort group record pointer
               iyy3 = (nc.gtot - nc.stot) * (ixx2 and (nc.rcnt = nc.ntot))
               iyy4 = (nc.gtot - nc.stot) * (ixx1 and (nc.rcnt = nc.ntot))
               lrcd = clng(nc.rcnt + ixx1) * iyy2 + iyy3 * iyy1 + ixx4 - ixx1
               lrcx = clng(nc.rcnt + ixx2) * iyy2 + iyy4 * iyy1 + ixx3 - ixx2
               if nc.mndx then            'get index pointer from memory buffer
                  mid$(cnc.lng, 1) = mid$(cnc.nbuf, cint(lrcd), 3)'get f/memory
               else                                'get index pointer from file
                  i = ips.rget(nc.ndno, nc.nlen, lrcd, cnc.nbuf) 'get from file
                  mid$(cnc.lng, 1) = cnc.nbuf   'put file value to long-integer
               end if              'add one record from each sort group [below]
               i = ips.rget(nc.sdno, nc.slen, cvl(cnc.lng), cnc.sdat) 'get data
               i = ips.fill(nc, cnc.data, cnc.indx, cnc.lng, _ 'add to the sort
               cnc.sdat, insbeg(), inslen(), insseq())
               lnsptr(nc.rcnt - 1, 0) = lrcd  'begin ptr.for group index record
               lnsptr(nc.rcnt - 1, 1) = lrcx    'end ptr.for group index record
            next
         end if
      end if
      cnum = ltrim$(str$(nc.rtot))          'character form of word-total value
      locate nc.drow - 3, 1, 0           'position cursor for following message
      print " No. 2 of 3 processes (3rd process is faster)";      'User message
      locate nc.drow - 1, 1, 0           'position cursor for following message
      print " Writing "; space$(len(cnum)); " of "; cnum; " pointers";
      icncptr = 1                        'initialize the pointer-output pointer
      icncrcd = 0                    'initialize pointer-output "record" number
      if nc.ntot then                         'merge-retrieval from sort groups
         do while nc.rcnt                     'sort records are still available
            lset cnc.lng = cnc.indx       'get pointer-to-index from work-index
            mid$(cnc.lng, 4) = char(0)          'use only the first three bytes
            nc.gptr = cvl(cnc.lng)                 'set the group-pointer value
            iptr = nc.gptr - 1                    'temp value used for lnsptr()
            mid$(cnc.indx, 1) = mid$(cnc.indx, 4)         'shift the work-index
            if nc.mndx then            'retrieval index fits entirely in memory
               mid$(cnc.ptr, icncptr) = mid$(cnc.nbuf, cint(lnsptr(iptr, 0)), 3)
            else                        'retrieval index does not fit in memory
               i = ips.rget(nc.ndno, nc.nlen, lnsptr(iptr, 0), cnc.nbuf)
               mid$(cnc.ptr, icncptr) = cnc.nbuf 'save in pointer-output buffer
            end if
            if lnsptr(iptr, 0) = lnsptr(iptr, 1) then   'end of this sort group
               nc.rcnt = nc.rcnt - 1             'decrement group stack pointer
            else     'end of group not yet reached; increment the group pointer
               lnsptr(iptr, 0) = lnsptr(iptr, 0) + iyy1 * nc.sseq
               if nc.mndx then         'retrieval index fits entirely in memory
                  mid$(cnc.lng, 1) = mid$(cnc.nbuf, cint(lnsptr(iptr, 0)), 3)
               else                     'retrieval index does not fit in memory
                  i = ips.rget(nc.ndno, nc.nlen, lnsptr(iptr, 0), cnc.nbuf)
                  mid$(cnc.lng, 1) = cnc.nbuf    'put pointer into long-integer
               end if
               i = ips.rget(nc.sdno, nc.slen, cvl(cnc.lng), cnc.sdat) 'get data
               i = ips.fill(nc, cnc.data, cnc.indx, cnc.lng, _     '...to merge
               cnc.sdat, insbeg(), inslen(), insseq())       '...into the stack
            end if
            icncptr = icncptr + 3         'increment the pointer-output pointer
            if icncptr > iptrlen or nc.rcnt = 0 then'buffer full or end-of-sort
               if nc.rcnt = 0 then                  'sort retrieval is complete
                  ircdtot = icncrcd + 1           'total pointer-buffer records
                  isublen = icncptr - 1          'length of final output buffer
                  cnc.ptr = left$(cnc.ptr, isublen)   'resize the output buffer
               end if
               icncptr = 1                    'reset the pointer-output pointer
               icncrcd = icncrcd + 1     'increment pointer-output "record" no.
               i = ips.rput(nc.ddno, iptrlen, icncrcd, cnc.ptr)'output pointers
               locate nc.drow - 1, 10, 0 'position cursor for following message
               print ltrim$(str$(clng(icncrcd) * 100));  'current counter value
            end if
         loop
         locate nc.drow - 3, 1, 0        'position cursor for following message
         print " Third and final process.....                ";   'User message
         locate nc.drow - 1, 1, 0        'position cursor for following message
         print " Converting "; space$(len(cnum)); " of "; cnum; " pointers";
         cnc.ptr = space$(iptrlen)      'reinitialize the pointer-output buffer
         for icncrcd = 1 to ircdtot        'loop thru the pointer-output groups
            if icncrcd = ircdtot then              'current group = final group
               cnc.ptr = space$(isublen)              'resize the output buffer
            end if
            i = ips.rget(nc.ddno, iptrlen, clng(icncrcd), cnc.ptr) 'load buffer
            for icncptr = 1 to len(cnc.ptr) step 3'loop thru the group-pointers
               mid$(cnc.lng, 1) = mid$(cnc.ptr, icncptr, 3)'put pointer to long
               i = ips.rget(nc.sdno, nc.slen, cvl(cnc.lng), cnc.sdat)'load data
               mid$(cnc.ptr, icncptr + 0) = char(midchar(cnc.sdat, nc.slen - 0))
               mid$(cnc.ptr, icncptr + 1) = char(midchar(cnc.sdat, nc.slen - 1))
               mid$(cnc.ptr, icncptr + 2) = char(midchar(cnc.sdat, nc.slen - 2))
            next           'replace pointer-to-data with actual pointer [above]
            i = ips.rput(nc.ddno, iptrlen, icncrcd, cnc.ptr)   'save the buffer
            locate nc.drow - 1, 13, 0    'position cursor for following message
            print ltrim$(str$(clng(icncrcd) * 100));     'current counter value
         next
      else                     'non-merge sort retrieval; all data is in memory
         do while nc.dtot                     'sort records are still available
            mid$(cnc.lng, 1) = mid$(cnc.nbuf, nc.dptr * 3 - 2, 3)'long-int.buff
            i = ips.rget(nc.sdno, nc.slen, cvl(cnc.lng), cnc.sdat)'data&pointer
            mid$(cnc.ptr, icncptr + 0) = char(midchar(cnc.sdat, nc.slen - 0))
            mid$(cnc.ptr, icncptr + 1) = char(midchar(cnc.sdat, nc.slen - 1))
            mid$(cnc.ptr, icncptr + 2) = char(midchar(cnc.sdat, nc.slen - 2))
            nc.dptr = nc.dptr + nc.sseq   'increment or decrement stack pointer
            nc.dtot = nc.dtot - 1                  'decrement remaining records
            icncptr = icncptr + 3         'increment the pointer-output pointer
            if icncptr > iptrlen or nc.dtot = 0 then'buffer full or end of sort
               if nc.dtot = 0 then                  'sort retrieval is complete
                  cnc.ptr = left$(cnc.ptr, icncptr - 1)      'resize the buffer
               end if
               icncptr = 1                    'reset the pointer-output pointer
               icncrcd = icncrcd + 1  'increment pointer-output "record" number
               i = ips.rput(nc.ddno, iptrlen, icncrcd, cnc.ptr)'output pointers
            end if
         loop
      end if
      i = ifn.kill(nc.sdno, "sortdata.sdx")                'kill sort data file
      i = ifn.kill(nc.ndno, "sortdata.ndx")           'kill sort index workfile
      close nc.ddno                        'close the output pointer/index file
      cnc.data = ""                                'kill main sort group buffer
      cnc.indx = ""                                     'kill sort index buffer
      cnc.nbuf = ""                                     'kill index file buffer
      cnc.sdat = ""                                      'kill sort data buffer
   end if
end function


function ips.fill(nc as ncnf, cnc.data, cnc.indx, _      'add data to the stack
cnc.lng, cnc.sdat, insbeg(), inslen(), insseq())
   itop = nc.rcnt                                 'set top end of binary search
   ilow = 0                                       'set low end of binary search
   do while itop - ilow > 1    'search work data buffer using work index buffer
      imid = ilow + (itop - ilow) \ 2                'set mid point for compare
      mid$(cnc.lng, 1) = mid$(cnc.indx, imid * 3 - 2, 3)  'set long-int f/index
      iptr = (cvl(cnc.lng) - nc.offs - 1) * nc.slen + 1'pointer to index buffer
      if cnc.sdat <= mid$(cnc.data, iptr, nc.slen) then  'new data is <= buffer
         itop = imid                                         'move search lower
      else             'sort record value > compare value in sort memory buffer
         ilow = imid                                        'move search higher
      end if
   loop
   iptr = itop * 3 - 2                   'current index-"stack" insert position
   mid$(cnc.data, (nc.gptr - 1) * nc.slen + 1) = cnc.sdat      'write sort data
   mid$(cnc.indx, iptr + 3) = mid$(cnc.indx, iptr, (nc.rcnt - itop) * 3) 'shift
   mid$(cnc.indx, iptr) = left$(mkl$(nc.offs + nc.gptr), 3)
end function                            'write current pointer to index [above]


function ips.rget(ifno, ilen, lrcd, cbuf)           'get a record from the file
   get ifno, (lrcd - 1) * ilen + 1, cbuf                        'get the record
end function


function ips.rput(ifno, ilen, ircd, cbuf)             'put a record to the file
   put ifno, clng(ircd - 1) * ilen + 1, cbuf                    'put the record
end function


function ips.ssav(nc as ncnf, cnc.data, cnc.indx, insbeg(), inslen(), insseq())
   nc.ntot = nc.ntot + 1                      'increment the index group number
   i = ips.rput(nc.ndno, nc.nlen, nc.ntot, left$(cnc.indx, nc.nlen))
   ilen = nc.rmax * nc.slen                        'size of group memory buffer
   irec = lof(nc.sdno) \ ilen + 1                        'current data "record"
   i = ips.rput(nc.sdno, ilen, irec, cnc.data)          'put data group to file
end function


function lps.free(iexc, lmax)           'get maximum free memory less exclusion
   if fre("") - iexc > lmax then                  'available memory > requested
      lps.free = lmax                                 'return requested maximum
   else                                         'requested memory NOT available
      lps.free = fre("") - iexc              'return free memory less exclusion
   end if
end function
